"""Inbound message handler."""

from abc import ABCMeta

from ..subscriber_base import SubscriberBase
from ..utils import build_topic


class InboundHandlerBase(SubscriberBase):
    """Inbound message handler."""

    __meta__ = ABCMeta

    def __init__(self, topic, address=None, group=None, message_type=None):
        """Init the InboundHandlerBase class."""
        subscriber_topic = build_topic(
            prefix="handler",
            topic=topic,
            address=address,
            group=group,
            message_type=message_type,
        )
        self._topic = topic
        super().__init__(subscriber_topic=subscriber_topic)
        for attr_str in dir(self):
            attr = getattr(self, attr_str)
            if hasattr(attr, "register_topic"):
                attr.register_topic(
                    attr, topic, address=address, group=group, message_type=message_type
                )
            if hasattr(attr, "register_status") and hasattr(self, "_address"):
                # pylint: disable=no-member
                attr.register_status(attr, self._address.id)



"""Heartbeat manager."""

import asyncio
from datetime import datetime, timedelta

from ..address import Address
from ..handlers.from_device.off import OffInbound
from ..handlers.from_device.on_level import OnLevelInbound
from ..subscriber_base import SubscriberBase

HB_CHECK_BUFFER = 300  # 5 min or 300 seconds


class HeartbeatManager(SubscriberBase):
    """Heartbeat manager."""

    class OnOffHeartbeat(SubscriberBase):
        """On / Off events for subscribers."""

        def call_subscribers(self, on_level):
            """Call subscribers to the event."""
            self._call_subscribers(on_level=on_level)

    def __init__(self, address, group, max_duration=1275):
        """Init the HeartbeatManager class."""
        self._address = Address(address)
        self._group = group
        self._max_duration = max_duration
        subscriber_topic = "subscriber_{}_heartbeat".format(self._address.id)
        super().__init__(subscriber_topic)

        self._on_inbound = OnLevelInbound(self._address, self._group)
        self._off_inbound = OffInbound(self._address, self._group)
        self._on_inbound.subscribe(self._on_heartbeat_received)
        self._off_inbound.subscribe(self._off_heartbeat_received)

        self._on_hb = self.OnOffHeartbeat("{}_{}".format(subscriber_topic, "on"))
        self._off_hb = self.OnOffHeartbeat("{}_{}".format(subscriber_topic, "off"))

        self._last_heartbeat = datetime.now()
        self._schedule_next_check()

    @property
    def max_duration(self):
        """Return the maximum heartbeat duration."""
        return self._max_duration

    @max_duration.setter
    def max_duration(self, value):
        """Set the maximum heartbeat duration."""
        self._max_duration = float(value)

    def subscribe_on(self, callback):
        """Subscribe to ON heartbeat events."""
        self._on_hb.subscribe(callback)

    def subscribe_off(self, callback):
        """Subscribe to OFF heartbeat events."""
        self._off_hb.subscribe(callback)

    def _on_heartbeat_received(self, on_level):
        """Listen for ON messages from device."""
        self._last_heartbeat = datetime.now()
        self._call_subscribers(heartbeat=False)
        self._on_hb.call_subscribers(on_level=0xFF)

    def _off_heartbeat_received(self, on_level):
        """Listen for OFF messages from device."""
        self._last_heartbeat = datetime.now()
        self._call_subscribers(heartbeat=False)
        self._off_hb.call_subscribers(on_level=0)

    def _check_heartbeat(self):
        """Check if a heartbeat as arrived since max_duration."""
        td_last_heartbeat = datetime.now() - self._last_heartbeat
        if td_last_heartbeat > timedelta(minutes=self._max_duration):
            self._call_subscribers(heartbeat=True)
        else:
            self._call_subscribers(heartbeat=False)
        self._schedule_next_check()

    def _schedule_next_check(self):
        """Schedule the next time we check for the heartbeat."""
        loop = asyncio.get_event_loop()
        # Last heartbeat time is a baseline to trigger the next call
        last_hb = (datetime.now() - self._last_heartbeat).total_seconds()
        # Calculate seconds from now to check again
        # _max_duration is in minutes so convert to seconds
        max_dur_sec = self._max_duration * 60
        next_call = max(HB_CHECK_BUFFER, max_dur_sec + HB_CHECK_BUFFER - last_hb)
        loop.call_later(next_call, self._check_heartbeat)



"""Command line tools to interact with the Insteon devices."""
from binascii import unhexlify
import os
from .. import async_connect, async_close, devices
from ..constants import HC_LOOKUP, UC_LOOKUP
from .tools_base import ToolsBase
from .config import ToolsConfig
from .aldb import ToolsAldb
from .cmd import CmdTools
from ..address import Address
from ..managers.link_manager import async_enter_linking_mode, async_enter_unlinking_mode


class InsteonCmd(ToolsBase):
    """Command class to test interactivity."""

    async def do_connect(self, *args, **kwargs):
        """Connect to the Insteon modem.

        Usage:
            connect device
            connect host username [hub_version [port]]
        """
        args = args[0].split()
        try:
            self.username = args[1]
            self.host = args[0]
        except IndexError:
            try:
                self.device = args[0]
            except IndexError:
                pass
            params = self.device
        else:
            try:
                self.hub_version = int(args[2])
                self.port = int(args[3])
            except (ValueError, IndexError):
                self.hub_version = None
                self.port = None

        password = await self._get_connection_params()

        if password:
            params = f"{self.host} {self.username} {'*' * len(password)} {self.hub_version} {self.port}"
        self._log_command(f"connect {params}")
        try:
            await async_connect(
                device=self.device,
                host=self.host,
                port=self.port,
                username=self.username,
                password=password,
                hub_version=self.hub_version,
            )
        except ConnectionError:
            self._log_stdout("Connection failed. Please review connection information.")

    async def do_disconnect(self, *args, **kwargs):
        """Close the connection to the modem.

        Usage:
            disconnect
        """
        self._log_command("disconnect")
        await async_close()

    async def do_load_devices(self, *args, **kwargs):
        """Load the devices.

        Usage:
            load_devices workdir id_devices load_modem_aldb

        workdir: Directory where the saved device file is located (Enter . for current directory)
        id_devices: Option for handling unknown devices
            0 - Do not ID devices
            1 - ID unknown devices only (default)
            2 - ID all devices
        load_modem_aldb: Option for loading the Modem ALDB
            0 - Do not load
            1 - Load if not loaded from saved file (default)
            2 - Always load
        """

        args = args[0].split()
        try:
            if args[0] != "":
                self.workdir = args[0]
                if self.workdir == ".":
                    self.workdir = os.getcwd()
        except IndexError:
            pass

        if not self.workdir:
            self.workdir = await self._get_workdir()

        try:
            id_devices = int(args[1])
        except (IndexError, ValueError):
            id_devices = None
        if id_devices not in [0, 1, 2]:
            id_devices = await self._get_int(
                "Identify devices (0=None, 1=Unknown Only, 2=All", 1, [0, 1, 2]
            )
        try:
            load_modem_aldb = int(args[2])
        except (IndexError, ValueError):
            load_modem_aldb = None
        if load_modem_aldb not in [0, 1, 2]:
            load_modem_aldb = await self._get_int(
                "Identify devices (0=No, 1=If no loaded, 2=Load", 1, [0, 1, 2]
            )
        self._log_command(f"load_devices {self.workdir} {id_devices}")
        await devices.async_load(
            workdir=self.workdir, id_devices=id_devices, load_modem_aldb=load_modem_aldb
        )
        self._log_stdout(f"Total devices: {len(devices)}")

    async def do_manage_aldb(self, *args, **kwargs):
        """Manage device All-Link database."""
        self._log_command("manage_aldb")
        await self._call_next_menu(ToolsAldb, "aldb")

    async def do_manage_config(self, *args, **kwargs):
        """Manage operational flags and extended properties."""
        self._log_command("manage_config")
        await self._call_next_menu(ToolsConfig, "config")

    async def do_commands(self, *args, **kwargs):
        """Execute device commands."""
        self._log_command("tests")
        await self._call_next_menu(CmdTools, "commands")

    async def do_link_device(self, *args, **kwargs):
        """Link a device to the modem."""
        self._log_command("add_device")
        self._log_stdout(
            "Press the set button on the device. Linking will occur in the background."
        )
        await async_enter_linking_mode(is_controller=True, group=0)

    async def do_add_device_manually(self, *args, **kwargs):
        """Add a device using a cat and subcat.

        Usage:
           add_device_manually address cat subcat [firmware]

        address: Device address
        cat: Device category (i.e. 05 or 0x05)
        subcat: Device subcategory (i.e. 1f or 0x1f)

        Note: Device cat and subcat are the hex representation and should be at
        least 2 digits. They can be written with or without the hex prefix 0x.

        The cat and subcat can be found in the user manual for the device. The values
        in the user manual are in hex.
        """

        def str_to_hex(value):
            """Convert a hex string into an integer."""
            if value[0:2] == "0x":
                value = value[2:]
            return unhexlify(value)

        args = args[0].split()
        try:
            address = Address(args[0])
        except (IndexError, ValueError):
            address = None

        try:
            cat = str_to_hex(args[1])
        except (IndexError, ValueError):
            cat = None

        try:
            subcat = str_to_hex(args[2])
        except (IndexError, ValueError):
            subcat = None

        try:
            firmware = str_to_hex(args[3])
        except (IndexError, ValueError):
            firmware = 0

        if address is None:
            address = await self._get_char("Enter device address (i.e. 1a2b3c")
            try:
                address = Address(address)
            except ValueError:
                self._log_stdout("Invalid address")
                return

        if cat is None:
            cat = await self._get_char("Enter device cat (i.e. 10 or 0x10")
            try:
                cat = str_to_hex(cat)
            except ValueError:
                self._log_stdout("Invalid device category")
                return

        if subcat is None:
            subcat = await self._get_char("Enter device cat (i.e. 10 or 0x10")
            try:
                subcat = str_to_hex(subcat)
            except ValueError:
                self._log_stdout("Invalid device subcategory")
                return

        if firmware is None:
            firmware = await self._get_char("Enter device cat (i.e. 10 or 0x10", "00")
            try:
                firmware = str_to_hex(firmware)
            except ValueError:
                self._log_stdout("Invalid device firmware")
                return

        devices.set_id(address, cat, subcat, firmware)

    async def do_remove_device(self, *args, **kwargs):
        """Add a device."""
        self._log_command("remove_device")
        self._log_stdout(
            "Press the set button on the device. Unlinking will occur in the background."
        )
        await async_enter_unlinking_mode(group=0)

    async def do_exit(self, *args, **kwargs):
        """Exit the current menu.

        Usage:
            exit
        """
        self._log_command("exit")
        await self.do_log_to_file("n")
        return -1

    async def do_add_x10_device(self, *args, **kwargs):
        """Add an X10 device.

        Usage:
            add_x10_device housecode unitcode type [steps]

        houscode: The device housecode (a - p)
        unitcode: The device unitcode (1 - 16)
        type: The device type (on_off, dimmable, sensor)
        steps: Number of steps from off to full on (dimmable only)
        """
        args = args[0].split()
        x10_types = ["dimmable", "on_off", "sensor"]
        try:
            housecode = args[0]
            if housecode not in list(HC_LOOKUP.keys()):
                housecode = None
        except IndexError:
            housecode = None

        try:
            unitcode = int(args[1])
            if unitcode not in list(UC_LOOKUP.keys()):
                unitcode = None
        except (IndexError, ValueError):
            unitcode = None

        try:
            x10_type = args[2]
            if x10_type not in x10_types:
                x10_type = None
            try:
                steps = int(args[3])
            except IndexError:
                steps = 22
            except ValueError:
                steps = None
        except IndexError:
            x10_type = None
            steps = None

        if housecode is None:
            housecode = await self._get_char(
                "Enter housecode", values=list(HC_LOOKUP.keys())
            )

        if unitcode is None:
            unitcode = await self._get_int(
                "Enter unitcode", values=list(UC_LOOKUP.keys())
            )

        if x10_type is None:
            x10_type = await self._get_char("Enter type", values=x10_type)

        if x10_type == "dimmable" and steps is None:
            steps = await self._get_int("Dimmer steps", default=22)

        devices.add_x10_device(housecode, unitcode, x10_type, steps, 255)


def tools():
    """Start insteon tools."""
    InsteonCmd.start()



"""Command line interface for Operation Flag and Extended Properites Management."""

from .tools_base import ToolsBase
from .. import devices
from ..constants import ResponseStatus, RelayMode


class ToolsConfig(ToolsBase):
    """Command line interface for Operation Flag and Extended Properites Management."""

    async def do_read_config(self, *args, **kwargs):
        """Read the operating flags and extended properties of a device.

        Usage:
            read_config <ADDRESS>|all
        """
        args = args[0].split()
        try:
            address = args[0]
        except IndexError:
            address = None

        addresses = await self._get_addresses(
            address=address, allow_cancel=True, allow_all=True
        )
        if not addresses:
            return

        for address in addresses:
            device = devices[address]
            if not device == devices.modem:
                self._log_stdout(f"Reading configuration for: {device.address}")
                if await device.async_read_op_flags() == ResponseStatus.SUCCESS:
                    self._log_stdout("Operating flags read")
                else:
                    self._log_stdout("Operating flags read error")
                if await device.async_read_ext_properties() == ResponseStatus.SUCCESS:
                    self._log_stdout("Extended properties read")
                else:
                    self._log_stdout("Extended properties read error")

    async def do_update_config(self, *args, **kwargs):
        """Write the operating flags and extended properties to a device.

        Usage:
            update_config <ADDRESS>|all
        """
        args = args[0].split()
        try:
            address = args[0]
        except IndexError:
            address = None

        addresses = await self._get_addresses(
            address, allow_cancel=True, allow_all=True
        )
        if not addresses:
            return

        for address in addresses:
            device = devices[address]
            if not device == devices.modem:
                if await device.async_write_op_flags() == ResponseStatus.SUCCESS:
                    self._log_stdout("Operating flags writen successfully")
                else:
                    self._log_stdout("Operating flags not writen")
                if await device.async_write_ext_properties() == ResponseStatus:
                    self._log_stdout("Extended properties writen successfully")
                else:
                    self._log_stdout("Extended properties not writen")

    async def do_print_config(self, *args, **kwargs):
        """Write the operating flags and extended properties to a device.

        Usage:
            read_ops_flags <ADDRESS>|all
        """
        args = args[0].split()
        try:
            address = args[0]
        except IndexError:
            address = None

        addresses = await self._get_addresses(
            address, allow_cancel=True, allow_all=True
        )
        if not addresses:
            return

        for address in addresses:
            device = devices[address]
            if device and not device == devices.modem:
                self._log_stdout("")
                self._log_stdout(f"Device: {address}")
                self._log_stdout("Operating Flag                  Value")
                self._log_stdout("------------------------------  -----")
                for name in device.operating_flags:
                    op_flag = device.operating_flags[name]
                    if op_flag.is_dirty:
                        name_out = f"{name}*"
                    else:
                        name_out = name
                    self._log_stdout(f"{name_out:30s}  {bool(op_flag.value)}")
                self._log_stdout("")
                self._log_stdout("Property                        Value")
                self._log_stdout("------------------------------  ----------")
                for name in device.properties:
                    prop = device.properties[name]
                    if prop.is_dirty:
                        name_out = f"{name}*"
                    else:
                        name_out = name
                    if isinstance(prop.value, bool):
                        self._log_stdout(f"{name_out:30s}  {prop.value}")
                    else:
                        prop_value = prop.value if prop.value is not None else 0
                        self._log_stdout(
                            f"{name_out:30s}  0x{prop_value:02x} ({prop_value:d})"
                        )
                self._log_stdout("")

    async def do_set_config_value(self, *args, **kwargs):
        """Set an operating flag or extended property for a device.

        Usage:
            set_config_value <ADDRESS> flag y|n
        """
        args = args[0].split()
        try:
            address = args[0]
        except IndexError:
            address = None

        addresses = await self._get_addresses(
            address, allow_cancel=True, allow_all=False
        )
        if not addresses:
            return
        device = devices[addresses[0]]

        try:
            name = args[1]
        except IndexError:
            name = None

        if name is None:
            name = await self._get_char("Enter property name")

        if not name:
            self._log_stdout("A name is required")

        if not device.operating_flags.get(name) and not device.properties.get(name):
            self._log_stdout(f"Flag {name} not found in device {addresses[0]}")
            return

        if device.operating_flags.get(name):
            prop = device.operating_flags[name]
        else:
            prop = device.properties[name]

        try:
            value = args[2]
            if isinstance(prop.value, bool):
                if value not in ["y", "n"]:
                    self._log_stdout("Value must be y or n")
                    return
                value = value == "y"
            else:
                value = int(value)
                if value not in range(0, 256):
                    self._log_stdout("Value must be between 0 and 255")
                    return
        except IndexError:
            value = None
        except TypeError:
            self._log_stdout("Value must be an integer between 0 and 255")
            return

        if value is None:
            if isinstance(prop.value, bool):
                yn_value = await self._get_char(
                    "Property is set (y/n)", values=["y", "n"]
                )
                value = yn_value == "y"
            else:
                value = await self._get_int("Property value", values=range(0, 256))

        prop.new_value = value
        self._log_stdout(f"Value {value} writen to property {prop.name}")

    async def do_write_config(self, *args, **kwargs):
        """Write the device config.

        Usage:
            write_config <ADDRESS>|all
        """
        args = args[0].split()
        try:
            address = args[0]
        except IndexError:
            address = None

        addresses = await self._get_addresses(
            address, allow_cancel=True, allow_all=True
        )
        if not addresses:
            return

        for address in addresses:
            device = devices[address]
            if device != devices.modem:
                if await device.async_write_op_flags() == ResponseStatus.SUCCESS:
                    self._log_stdout("Operating flags written")
                else:
                    self._log_stdout("Operating flags write error")
                if await device.async_write_ext_properties() == ResponseStatus.SUCCESS:
                    self._log_stdout("Extended properties written")
                else:
                    self._log_stdout("Extended properties write error")

    async def do_set_kpl_toggle_mode(self, *args, **kwargs):
        """Set the toggle mode of a KeypadLinc button.

        Usage:
            set_kpl_toggle_mode address button mode

        address: Address of a KeypadLinc device
        button: Button number to set (2-8 on an 8 button KPL, 3-6 on a 6 button KPL)
        mode: Button mode (0: toggle on/off, 1: On only, 2: Off only)
        """
        args = args[0].split()
        try:
            address = args[0]
        except IndexError:
            address = None

        addresses = await self._get_addresses(
            address, allow_cancel=True, allow_all=False
        )
        if not addresses:
            return

        device = devices[addresses[0]]
        if not hasattr(device, "async_set_toggle_mode"):
            self._log_stdout("Device is not a KeypadLinc")
            return

        try:
            button = int(args[1])
        except (IndexError, ValueError):
            button = await self._get_int("Buttom number", values=device.groups.keys())

        try:
            mode = int(args[2])
        except (IndexError, ValueError):
            mode = await self._get_int(
                "Toggle mode (0: toggle on/off, 1: On only, 2: Off only)",
                values=[0, 1, 2],
            )
        await device.async_set_toggle_mode(button=button, mode=mode)

    async def do_set_kpl_radio_buttons(self, *args, **kwargs):
        """Set the toggle mode of a KeypadLinc button.

        Usage:
            set_kpl_radio_buttons address button1 button2 [button3 button4 button5 button6 button7]

        address: Address of a KeypadLinc device
        button: Button number to set (2-8 on an 8 button KPL, 3-6 on a 6 button KPL)
        """
        args = args[0].split()
        try:
            address = args[0]
        except IndexError:
            address = None

        addresses = await self._get_addresses(
            address, allow_cancel=True, allow_all=False
        )
        if not addresses:
            return

        device = devices[addresses[0]]
        if not hasattr(device, "async_set_radio_buttons"):
            self._log_stdout("Device is not a KeypadLinc")
            return

        item = 1
        buttons = []
        while True:
            try:
                buttons.append(int(args[item]))
                item += 1
            except (IndexError, ValueError):
                break

        possible_buttons = [0]
        _ = [possible_buttons.append(k) for k in device.groups]
        if len(buttons) < 2:
            while True:
                for button in buttons:
                    if button in possible_buttons:
                        possible_buttons.remove(button)
                button = await self._get_int(
                    "Button number (enter 0 for last button)", values=possible_buttons
                )
                if button != 0:
                    buttons.append(button)
                else:
                    break
        if len(buttons) < 2:
            self._log_stdout("At least two buttons are required")
            return

        await device.async_set_radio_buttons(buttons=buttons)

    async def do_set_iolinc_mode(self, *args, **kwargs):
        """Set the mode of an IO Linc device.

        Usage:
            set_iolinc_mode address mode

        address: IOLinc device address
        mode: Trigger or Momentary A, B or C
            0 - Latching (Continuous)
            1 - Momentary A
            2 - Momentary B
            3 - Momentary C

        for more information see the IOLinc user manual.
        """
        args = args[0].split()
        try:
            address = args[0]
        except IndexError:
            address = None

        addresses = await self._get_addresses(
            address, allow_cancel=True, allow_all=False
        )
        if not addresses:
            return

        device = devices[addresses[0]]
        if not hasattr(device, "async_set_relay_mode"):
            self._log_stdout("Device is not an IOLinc device")
            return

        try:
            mode = int(args[1])
            if mode not in [0, 1, 2, 3]:
                raise ValueError
        except (IndexError, ValueError):
            mode = await self._get_int("Mode", values=[0, 1, 2, 3])

        await device.async_set_relay_mode(mode=RelayMode(mode))

    async def do_set_iolinc_delay(self, *args, **kwargs):
        """Set the delay of an IO Linc device when in momentary mode.

        Usage:
            set_iolinc_delay address seconds

        address: IOLinc device address
        seconds: Integer number of seconds to delay before the relay is turned off

        For more information see the IOLinc user manual.
        """
        args = args[0].split()
        try:
            address = args[0]
        except IndexError:
            address = None

        addresses = await self._get_addresses(
            address, allow_cancel=True, allow_all=False
        )
        if not addresses:
            return

        device = devices[addresses[0]]
        if not hasattr(device, "async_set_momentary_delay"):
            self._log_stdout("Device is not an IOLinc device")
            return

        try:
            seconds = int(args[1])
        except (IndexError, ValueError):
            seconds = await self._get_int("Delay (seconds)")

        await device.async_set_momentary_delay(seconds=seconds)

    async def do_get_engine_version(self, *args, **kwargs):
        """Get the engine version of the device.

        Usage:
            get_engine_version <ADDRESS>|all
        """
        args = args[0].split()
        try:
            address = args[0]
        except IndexError:
            address = None

        addresses = await self._get_addresses(
            address, allow_cancel=True, allow_all=True
        )
        if not addresses:
            return

        for address in addresses:
            device = devices[address]
            if not device == devices.modem and device.cat != 0x03:
                await device.async_get_engine_version()



"""Event class used to trigger event actions to subscribers.

Events include device triggered actions such as:
    - ON
    - OFF
    - FAST ON
    - FAST OFF

Subscribers to Event use the following interface:
    subscriber(name, address)
Where name is the name of the event and address is the address of the device
triggering the event.
"""
from .address import Address
from .subscriber_base import SubscriberBase

ON_EVENT = "on_event"
ON_FAST_EVENT = "on_fast_event"
OFF_EVENT = "off_event"
OFF_FAST_EVENT = "off_fast_event"
LOW_BATTERY_EVENT = "low_battery_event"
HEARTBEAT_EVENT = "heartbeat_event"
OPEN_EVENT = "open_event"
CLOSE_EVENT = "close_event"
LEAK_DRY_EVENT = "leak_dry_event"
LEAK_WET_EVENT = "leak_wet_event"
MOTION_DETECTED_EVENT = "motion_detected_event"
MOTION_TIMEOUT_EVENT = "motion_timeout_event"
LIGHT_DETECTED_EVENT = "light_detected_event"
DARK_DETECTED_EVENT = "dark_detected_event"
SMOKE_DETECTED_EVENT = "smoke_detected_event"
CO_DETECTED_EVENT = "co_detected_event"
TEST_DETECTED_EVENT = "test_detected_event"
NEW_DETECTED_EVENT = "new_detected_event"
LOW_BATTERY_EVENT = "low_battery_event"
SENSOR_MALFUNCTION_EVENT = "sensor_malfunction_event"
ALL_CLEAR_EVENT = "all_clear_event"


class Event(SubscriberBase):
    """Event class to manage triggering of events.

    Subscribers to Event use the following interface:

        subscriber(name, address, group)

    Where name is the name of the event and address is the address of the device
    triggering the event.
    """

    def __init__(self, name: str, address: Address, group=0, button=""):
        """Init the Event class."""
        self._address = address  # Address(address)
        self._group = group
        self._name = name
        self._button = button
        topic = "event_{}_{}_{}".format(self._address.id, group, name)
        super().__init__(subscriber_topic=topic)

    @property
    def address(self):
        """Address of the device sending the event."""
        return self._address

    @property
    def group(self):
        """Group number of the event."""
        return self._group

    @property
    def name(self):
        """Name of the event."""
        return self._name

    @property
    def button(self):
        """Return the button name."""
        return self._button

    def trigger(self, on_level):
        """Trigger the event."""
        self._call_subscribers(
            name=self._name,
            address=self._address.id,
            group=self._group,
            button=self._button,
        )


class LowBatteryEvent(Event):
    """Low battery event."""

    # pylint: disable=arguments-differ
    def trigger(self, low_battery):
        """Trigger the event."""
        self._call_subscribers(
            name=self._name, address=self._address.id, group=self._group
        )


class HeartbeatEvent(Event):
    """Heartbeat event."""

    # pylint: disable=arguments-differ
    def trigger(self, heartbeat):
        """Trigger the event."""
        self._call_subscribers(
            name=self._name, address=self._address.id, group=self._group
        )


class WetDryEvent(Event):
    """We or Dry event."""

    # pylint: disable=arguments-differ
    def trigger(self, dry):
        """Trigger the event."""
        self._call_subscribers(
            name=self._name, address=self._address.id, group=self._group
        )



"""Get Device Info command handler."""
from ...address import Address
from ...topics import ID_REQUEST
from .direct_command import DirectCommandHandlerBase


class IdRequestCommand(DirectCommandHandlerBase):
    """Get Device ID command handler."""

    def __init__(self, address: Address):
        """Init the IdRequest class."""
        super().__init__(topic=ID_REQUEST, address=address)

    # pylint: disable=arguments-differ
    async def async_send(self):
        """Send the device ID request asyncronously."""
        return await super().async_send()



"""Protocol classes to interface with serial, socket and http devices."""
import asyncio
import logging
from functools import partial

from ..handlers.get_im_info import GetImInfoHandler
from .http_transport import async_connect_http
from .protocol import Protocol
from .serial_transport import (
    async_connect_serial,
    async_connect_socket,
)
from ..managers.utils import create_device
from ..managers.device_id_manager import DeviceId

_LOGGER = logging.getLogger(__name__)


async def async_modem_connect(
    device=None, host=None, port=None, username=None, password=None, hub_version=2
):
    """Connect to the Insteon Modem.

    Returns an Insteon Modem object (PLM, Hub, or Hub1)

    Parameters:
        device: Serial port device (i.e. /dev/ttyUSB0 or COM5)
        host: Hub IP address (i.e. 192.168.1.100)
        port: Hub port number (Default 25105 for version 2 or 9761 for version 1)
        username: Hub username for the Hub V2
        password: Hub password for the Hub V2
        hub_version: 1 | 2 (Default: 2)

    If the device is a serial device see the serial class parameters.

    """
    device_id = None

    def set_im_info(address, cat, subcat, firmware):
        nonlocal device_id
        device_id = DeviceId(address, cat, subcat, firmware)

    async def async_test_device_id():
        """Test if the device ID is set."""
        nonlocal device_id
        retries = 120
        while device_id is None and retries:
            await asyncio.sleep(0.5)
            if device_id is not None:
                return True
            retries -= 1
        return False

    transport = None
    if not device and not host:
        return ValueError("Must specify either a device or a host")

    if device:
        connect_method = partial(async_connect_serial, **{"device": device})
        protocol = Protocol(connect_method=connect_method)

    elif hub_version == 2:
        connect_method = partial(
            async_connect_http,
            **{"host": host, "username": username, "password": password, "port": port},
        )
        protocol = Protocol(connect_method=connect_method)

    else:
        connect_method = partial(async_connect_socket, **{"host": host, "port": port})
        protocol = Protocol(connect_method=connect_method)

    try:
        await protocol.async_connect(retry=False)
    except ConnectionError:
        raise ConnectionError("Modem did not respond connection request")

    get_im_info = GetImInfoHandler()
    get_im_info.subscribe(set_im_info)
    await get_im_info.async_send()
    # Wait for a max of 60 seconds for the modem to respond
    if device_id is None and not await async_test_device_id():
        raise ConnectionError("Modem did not respond to ID request")
    modem = create_device(device_id)
    modem.protocol = protocol
    modem.transport = transport
    return modem



"""Base class for all group value entities."""

from abc import ABC, abstractmethod
from ..address import Address
from ..subscriber_base import SubscriberBase


class GroupBase(SubscriberBase):
    """Device group base class."""

    __meta__ = ABC

    def __init__(
        self, name: str, address: Address, group=0, default=None, value_type: type = int
    ):
        """Init the StateBase class."""
        self._address = address
        topic = "state_{}_{}_{}".format(self._address.id, name, group)
        super().__init__(subscriber_topic=topic)
        self._name = name
        self._group = group
        self._value = int(default) if default is not None else None
        self._type = value_type

    @property
    def name(self):
        """Return the name of the state."""
        return self._name

    @property
    def group(self):
        """Return the group number."""
        return self._group

    @property
    def value(self):
        """Return the value of the state."""
        return self._value

    @value.setter
    def value(self, value):
        """Set the value of the state."""
        try:
            value = self._type(value) if value is not None else None
            if self._value == value:
                return
            self._value = value
        except TypeError:
            raise TypeError(
                "Error setting value of State {}: "
                "Must be of type {}".format(self._name, self._type.__name__)
            )
        else:
            self._call_subscribers(
                name=self._name,
                address=self._address.id,
                value=self._value,
                group=self._group,
            )

    @abstractmethod
    def set_value(self, **kwargs):
        """Set the value of the state from a Handler."""



"""Wrapper for serial.aio transport."""
import asyncio
import logging

import serial
from serial_asyncio import SerialTransport as SerialTransportBase

_LOGGER = logging.getLogger(__name__)


async def async_connect_serial(device, protocol):
    """Connect to the PowerLine Modem via serial port.

    Parameters:
        port – Device name.
        protocol - Insteon Modem Protocol instance.

    """
    loop = asyncio.get_event_loop()
    try:
        ser = serial.serial_for_url(url=device, baudrate=19200)
        transport = SerialTransport(loop, protocol, ser, device=device)
    except OSError as ex:
        _LOGGER.warning("Unable to connect to %s: %s", device, ex)
        transport = None
    return transport


async def async_connect_socket(host, protocol, port=None):
    """Connect to the Hub Version 1 via TCP Socket."""

    port = 9761 if not port else port
    loop = asyncio.get_event_loop()
    url = "socket://{}:{}".format(host, port)
    try:
        ser = serial.serial_for_url(url=url, baudrate=19200)
        transport = SerialTransport(loop, protocol, ser, device=url)
    except OSError as ex:
        _LOGGER.warning("Unable to connect to %s: %s", url, ex)
        transport = None
    return transport


class SerialTransport(SerialTransportBase):
    """Wrapper for serial_asyncio.SerialTransport."""

    def __init__(self, loop, protocol, serial_instance, device):
        """Init the SerialTransport class."""
        super().__init__(loop=loop, protocol=protocol, serial_instance=serial_instance)
        self._device = device

    @property
    def connected(self):
        """Return True if the transport is connected to the serial device."""
        return not self._closing

    def write(self, data):
        """Override SerialTransport write method."""
        msg_bytes = bytes(data)
        try:
            super().write(msg_bytes)
        except serial.SerialException:
            _LOGGER.debug("Serial connection lost (write)")
            try:
                self._close()
            except serial.SerialException as exc:
                self._serial.close()
                self._protocol.connection_lost(exc)

    async def async_write(self, data):
        """Asyncronous write method."""
        self.write(data)

    async def async_connect(self, protocol, device=None, loop=None):
        """Connect to a serial device asycrounously."""
        device = device if device else self._device
        transport = async_connect_serial(device=device, protocol=protocol)
        return transport

    def _poll_read(self):
        try:
            super()._poll_read()
        except serial.SerialException:
            _LOGGER.debug("Serial connection lost (_poll_read)")
            self._close()

    def _call_connection_lost(self, exc):
        """Override _call_connection_lost in order to capture exceptions."""
        try:
            super()._call_connection_lost(exc)
        except serial.SerialException:
            _LOGGER.debug("Serial error (_call_connection_lost)")
            protocol = self._protocol
            self._write_buffer.clear()
            self._serial.close()
            self._serial = None
            self._protocol = None
            self._loop = None
            protocol.connection_lost(exc)



"""Python module for controlling Insteon devices."""
import asyncio
import logging
from pubsub import pub
from .protocol import async_modem_connect
from .managers.device_manager import DeviceManager
from .listener_exception_handler import ListenerExceptionHandler
from .handlers.from_device.x10_received import X10Received

# pylint: disable=unused-import
from .managers.x10_manager import (  # noqa: F401
    async_x10_all_lights_off,
    async_x10_all_lights_on,
    async_x10_all_units_off,
)

_LOGGER_TOPICS = logging.getLogger("pyinsteon.topics")
X10_RECEIVED_HANDLER = X10Received()

devices = DeviceManager()


async def async_connect(
    device=None,
    host=None,
    port=None,
    username=None,
    password=None,
    hub_version=2,
    **kwargs,
):
    """Connect to the Insteon Modem.

    Parameters:
        device: PLM serial / USB device name
        host: Hub hostname or IP address
        port: Hub TCP port
        username: Hub v2 username
        password: Hub v2 password
        hub_version: Hub version (default=2)

    Returns an Insteon Modem (PLM or Hub).

    """
    modem = await async_modem_connect(
        device=device,
        host=host,
        port=port,
        username=username,
        password=password,
        hub_version=hub_version,
        **kwargs,
    )
    devices.modem = modem
    devices.id_manager.start()
    return devices


async def async_close():
    """Close the connection and stop all tasks."""
    await devices.modem.async_close()
    for addr in devices:
        if devices[addr].is_battery:
            devices[addr].close()
    devices.id_manager.close()
    await asyncio.sleep(0.1)


def set_exception_handler():
    """Set the pubsub exception handler.

    This should only be run for debugging purposes.
    """
    pub.setListenerExcHandler(ListenerExceptionHandler())


def _log_all_topics(topic=pub.AUTO_TOPIC, **kwargs):
    """Log all topics from pyinsteon."""
    _LOGGER_TOPICS.debug("Topic: %s data: %s", topic.name, kwargs)


pub.subscribe(_log_all_topics, pub.ALL_TOPICS)



"""General controller devices (cat: 0x00)."""
from ..constants import ResponseStatus
from ..aldb.no_aldb import NoALDB
from ..operating_flag import (
    GROUPED_ON,
    KEY_BEEP_ON,
    LED_ON,
    PROGRAM_LOCK_ON,
    SEND_ON_ONLY,
    STAY_AWAKE_ON,
)
from ..groups import (
    ON_OFF_SWITCH_A,
    ON_OFF_SWITCH_B,
    ON_OFF_SWITCH_C,
    ON_OFF_SWITCH_D,
    ON_OFF_SWITCH_E,
    ON_OFF_SWITCH_F,
    ON_OFF_SWITCH_G,
    ON_OFF_SWITCH_H,
)
from .battery_base import BatteryDeviceBase
from .device_base import Device
from .variable_controller_base import VariableControllerBase


class GeneralController(Device):
    """General controller device."""

    def __init__(self, address, cat, subcat, firmware=0x00, description="", model=""):
        """Init the GeneralController class."""

        super().__init__(address, cat, subcat, firmware, description, model)
        self._aldb = NoALDB(self._address)


class GeneralController_ControlLinc(VariableControllerBase):
    """ControLinc 2430 device."""

    def __init__(self, address, cat, subcat, firmware=0x00, description="", model=""):
        """Init the GeneralController_ControlLinc class."""
        buttons = {
            1: ON_OFF_SWITCH_A,
            2: ON_OFF_SWITCH_B,
            3: ON_OFF_SWITCH_C,
            4: ON_OFF_SWITCH_D,
            5: ON_OFF_SWITCH_E,
            6: ON_OFF_SWITCH_F,
        }
        super().__init__(
            address, cat, subcat, firmware, description, model, buttons=buttons
        )


class GeneralController_RemoteLinc(BatteryDeviceBase, VariableControllerBase):
    """RemoteLinc 2440 device."""

    def __init__(self, address, cat, subcat, firmware=0x00, description="", model=""):
        """Init the GeneralController_RemoteLinc class."""
        buttons = {
            1: ON_OFF_SWITCH_A,
            2: ON_OFF_SWITCH_B,
            3: ON_OFF_SWITCH_C,
            4: ON_OFF_SWITCH_D,
            5: ON_OFF_SWITCH_E,
            6: ON_OFF_SWITCH_F,
        }
        super().__init__(
            address, cat, subcat, firmware, description, model, buttons=buttons
        )

    def _register_operating_flags(self):
        super()._register_operating_flags()
        self._add_operating_flag(STAY_AWAKE_ON, 0, 3, 6, 7)


class GeneralController_MiniRemoteBase(BatteryDeviceBase, VariableControllerBase):
    """RemoteLinc 2440 device."""

    def __init__(
        self,
        address,
        cat,
        subcat,
        firmware=0x00,
        description="",
        model="",
        buttons=None,
    ):
        """Init the GeneralController_MiniRemoteBase class."""
        super().__init__(
            address, cat, subcat, firmware, description, model, buttons=buttons
        )
        self._database_delta = 0

    async def async_status(self, group=0):
        """Return success always."""
        return ResponseStatus.SUCCESS

    def _register_operating_flags(self):
        super()._register_operating_flags()
        self._add_operating_flag(PROGRAM_LOCK_ON, 0, 0, 0, 1)
        self._add_operating_flag(LED_ON, 0, 1, 2, 3)
        self._add_operating_flag(KEY_BEEP_ON, 0, 2, 4, 5)
        self._add_operating_flag(STAY_AWAKE_ON, 0, 3, 6, 7)
        self._add_operating_flag(SEND_ON_ONLY, 0, 4, 8, 9)
        self._add_operating_flag(GROUPED_ON, 0, 6, 0x0E, 0x0F)


# TODO setup ability to swich from toggle to button groups
class GeneralController_MiniRemote_Switch(GeneralController_MiniRemoteBase):
    """RemoteLinc 2440 device with a single button."""

    def __init__(self, address, cat, subcat, firmware=0x00, description="", model=""):
        """Init the GeneralController_MiniRemote_Switch class."""
        buttons = {1: ON_OFF_SWITCH_A, 2: ON_OFF_SWITCH_B}
        super().__init__(
            address, cat, subcat, firmware, description, model, buttons=buttons
        )


class GeneralController_MiniRemote_4(GeneralController_MiniRemoteBase):
    """RemoteLinc 2440 device."""

    def __init__(self, address, cat, subcat, firmware=0x00, description="", model=""):
        """Init the GeneralController_MiniRemote_4 class."""
        buttons = {
            1: ON_OFF_SWITCH_A,
            2: ON_OFF_SWITCH_B,
            3: ON_OFF_SWITCH_C,
            4: ON_OFF_SWITCH_D,
        }
        super().__init__(
            address, cat, subcat, firmware, description, model, buttons=buttons
        )


class GeneralController_MiniRemote_8(GeneralController_MiniRemoteBase):
    """RemoteLinc 2440 device."""

    def __init__(self, address, cat, subcat, firmware=0x00, description="", model=""):
        """Init the GeneralController_MiniRemote_8 class."""
        buttons = {
            1: ON_OFF_SWITCH_B,
            2: ON_OFF_SWITCH_A,
            3: ON_OFF_SWITCH_D,
            4: ON_OFF_SWITCH_C,
            5: ON_OFF_SWITCH_F,
            6: ON_OFF_SWITCH_E,
            7: ON_OFF_SWITCH_H,
            8: ON_OFF_SWITCH_G,
        }
        super().__init__(
            address, cat, subcat, firmware, description, model, buttons=buttons
        )



"""INSTEON PLM constants for reuse across the module."""

from enum import Enum, IntEnum


class HexIntEnum(IntEnum):
    """IntEnum represented as a hex string or lowercase name."""

    def __repr__(self):
        """Emit the representation of the Enum."""
        return "0x{0:02x}".format(self.value)

    def __str__(self):
        """Emit the string of the Enum."""
        # pylint: disable=no-member
        return self.name.lower()


class DeviceCategory(HexIntEnum):
    """Device categories."""

    GENERALIZED_CONTROLLERS = 0x00
    DIMMABLE_LIGHTING_CONTROL = 0x01
    SWITCHED_LIGHTING_CONTROL = 0x02
    NETWORK_BRIDGES = 0x03
    IRRIGATION_CONTROL = 0x04
    CLIMATE_CONTROL = 0x05
    SENSORS_AND_ACTUATORS = 0x07
    HOME_ENTERTAINMENT = 0x08
    ENERGY_MANAGEMENT = 0x09
    BUILT_IN_APPLIANCE_CONTROL = 0x0A
    PLUMBING = 0x0B
    COMMUNICATION = 0x0C
    COMPUTER_CONTROL = 0x0D
    WINDOW_COVERINGS = 0x0E
    ACCESS_CONTROL = 0x0F
    SECURITY_HEALTH_SAFETY = 0x10
    SURVEILLANCE = 0x11
    AUTOMOTIVE = 0x12
    PET_CARE = 0x13
    TIMEKEEPING = 0x15
    HOLIDAY = 0x16


class MessageId(HexIntEnum):
    """Message type definitions."""

    STANDARD_RECEIVED = 0x50
    EXTENDED_RECEIVED = 0x51
    X10_RECEIVED = 0x52
    ALL_LINKING_COMPLETED = 0x53
    BUTTON_EVENT_REPORT = 0x54
    USER_RESET_DETECTED = 0x55
    ALL_LINK_CLEANUP_FAILURE_REPORT = 0x56
    ALL_LINK_RECORD_RESPONSE = 0x57
    ALL_LINK_CLEANUP_STATUS_REPORT = 0x58
    GET_IM_INFO = 0x60
    SEND_ALL_LINK_COMMAND = 0x61
    SEND_STANDARD = 0x62
    SEND_EXTENDED = 0x62
    X10_SEND = 0x63
    START_ALL_LINKING = 0x64
    CANCEL_ALL_LINKING = 0x65
    SET_HOST_DEV_CAT = 0x66
    RESET_IM = 0x67
    SET_ACK_MESSAGE_BYTE = 0x68
    GET_FIRST_ALL_LINK_RECORD = 0x69
    GET_NEXT_ALL_LINK_RECORD = 0x6A
    SET_IM_CONFIGURATION = 0x6B
    GET_ALL_LINK_RECORD_FOR_SENDER = 0x6C
    LED_ON = 0x6D
    LED_OFF = 0x6E
    MANAGE_ALL_LINK_RECORD = 0x6F
    SET_NAK_MESSAGE_BYTE = 0x70
    SET_ACK_MESSAGE_TWO_BYTES = 0x71
    RF_SLEEP = 0x72
    GET_IM_CONFIGURATION = 0x73


class MessageFlagType(HexIntEnum):
    """Message flag mesage type."""

    DIRECT = 0
    DIRECT_ACK = 1
    ALL_LINK_CLEANUP = 2
    ALL_LINK_CLEANUP_ACK = 3
    BROADCAST = 4
    DIRECT_NAK = 5
    ALL_LINK_BROADCAST = 6
    ALL_LINK_CLEANUP_NAK = 7


class AckNak(HexIntEnum):
    """ACK/NAK values."""

    ACK = 0x06
    NAK = 0x15


class ALDBStatus(HexIntEnum):
    """All-Link Database load status."""

    EMPTY = 0
    LOADED = 1
    LOADING = 2
    FAILED = 3
    PARTIAL = 4


class ALDBVersion(Enum):
    """All-Link Database version."""

    NULL = 0
    V1 = 1
    V2 = 2
    V2CS = 20


class EngineVersion(HexIntEnum):
    """Insteon Engine Version."""

    I1 = 0x00
    I2 = 0x01
    I2CS = 0x02
    UNKNOWN = 0x03


class X10CommandType(HexIntEnum):
    """X10 command types."""

    UNITCODE = 0x00
    COMMAND = 0x80


class ThermostatMode(HexIntEnum):
    """Thermostat system modes."""

    OFF = 0x00
    HEAT = 0x01
    COOL = 0x02
    AUTO = 0x03
    FAN_AUTO = 0x04
    FAN_ALWAYS_ON = 0x8


class FanSpeed(HexIntEnum):
    """Fan speeds."""

    OFF = 0x00
    LOW = 0x40
    MEDIUM = 0xC0
    HIGH = 0xFF


class FanSpeedRange(Enum):
    """Fan speed ranges."""

    OFF = [0x00]
    LOW = range(0x01, 0x7F)
    MEDIUM = range(0x80, 0xFE)
    HIGH = [0xFF]


class RelayMode(HexIntEnum):
    """Relay mode used by Sensor Actuator device class 0x07."""

    LATCHING = 0
    MOMENTARY_A = 1
    MOMENTARY_B = 2
    MOMENTARY_C = 3


class X10Commands(HexIntEnum):
    """X10 Commands."""

    ALL_UNITS_OFF = 0x00
    ALL_LIGHTS_ON = 0x01
    ALL_LIGHTS_OFF = 0x06
    ON = 0x02
    OFF = 0x03
    DIM = 0x04
    BRIGHT = 0x05
    EXTENDED_CODE = 0x07
    HAIL_REQUEST = 0x08
    HAIL_ACKNOWLEDGE = 0x09
    PRE_SET_DIM = 0x0A
    STATUS_IS_ON = 0x0B
    STATUS_IS_OFF = 0x0C
    STATUS_REQUEST = 0x0D


class ImButtonEvents(HexIntEnum):
    """IM Button Events values."""

    SET_TAPPED = 0x02
    SET_HELD = 0x03
    SET_RELEASED_AFTER_HOLD = 0x04
    SET_2_TAPPED = 0x12
    SET_2_HELD = 0x13
    SET_2_RELEASED_AFTER_HOLD = 0x14
    SET_3_TAPPED = 0x22
    SET_3_HELD = 0x23
    SET_3_RELEASED_AFTER_HOLD = 0x24


class AllLinkMode(HexIntEnum):
    """All Link Mode values."""

    RESPONDER = 0x00
    CONTROLLER = 0x01
    EITHER = 0x03
    DELETE = 0xFF


class ManageAllLinkRecordAction(HexIntEnum):
    """Manage All Link Record Action values."""

    FIND_FIRST = 0x00
    FIND_NEXT = 0x01
    MOD_FIRST_OR_ADD = 0x20
    MOD_FIRST_CTRL_OR_ADD = 0x40
    MOD_FIRST_RESP_OR_ADD = 0x41
    DELETE_FIRST = 0x80


class RampRate(IntEnum):
    """Ramp rate values."""

    MIN_9 = 0x00
    MIN_8 = 0x01
    MIN_7 = 0x02
    MIN_6 = 0x03
    MIN_5 = 0x04
    MIN_4_5 = 0x05
    MIN_4 = 0x06
    MIN_3_5 = 0x07
    MIN_3 = 0x08
    MIN_2_5 = 0x09
    MIN_2 = 0x0A
    MIN_1_5 = 0x0B
    MIN_1 = 0x0C
    SEC_47 = 0x0D
    SEC_43 = 0x0E
    SEC_38_5 = 0x0F
    SEC_34 = 0x10
    SEC_32 = 0x11
    SEC_30 = 0x12
    SEC_28 = 0x13
    SEC_26 = 0x14
    SEC_23_5 = 0x15
    SEC_21_5 = 0x16
    SEC_19 = 0x17
    SEC_8_5 = 0x18
    SEC_6_5 = 0x19
    SEC_4_5 = 0x1A
    SEC_2 = 0x1B
    SEC_0_5 = 0x1C
    SEC_0_3 = 0x1D
    SEC_0_2 = 0x1E
    SEC_0_1 = 0x1F


class ResponseStatus(HexIntEnum):
    """Response status of an outbound message."""

    FAILURE = 0
    SUCCESS = 1
    UNCLEAR = 2
    INPROGRESS = 4
    UNSENT = 8
    RUN_ON_WAKE = 0x10


class LinkStatus(Enum):
    """Status of a link between two devices."""

    MISSING_CONTROLLER = 0
    MISSING_RESPONDER = 1
    MISSING_TARGET = 2
    FOUND = 3
    TARGET_DB_NOT_LOADED = 4


MESSAGE_START_CODE = 0x02
MESSAGE_ACK = 0x06
MESSAGE_NAK = 0x15

CONTROL_FLAG_RECORD_IN_USE_0X80 = 0x80
CONTROL_FLAG_CONTROLER_0X40 = 0x40
CONTROL_FLAG_RECORD_HAS_BEEN_USED_0X02 = 0x02

# X10 House code lookup
HC_LOOKUP = {
    "a": 0x06,
    "b": 0x0E,
    "c": 0x02,
    "d": 0x0A,
    "e": 0x01,
    "f": 0x09,
    "g": 0x05,
    "h": 0x0D,
    "i": 0x07,
    "j": 0x0F,
    "k": 0x03,
    "l": 0x0B,
    "m": 0x00,
    "n": 0x08,
    "o": 0x04,
    "p": 0x0C,
}

# X10 Unit code lookup
UC_LOOKUP = {
    1: 0x06,
    2: 0x0E,
    3: 0x02,
    4: 0x0A,
    5: 0x01,
    6: 0x09,
    7: 0x05,
    8: 0x0D,
    9: 0x07,
    10: 0x0F,
    11: 0x03,
    12: 0x0B,
    13: 0x00,
    14: 0x08,
    15: 0x04,
    16: 0x0C,
}

RAMP_RATES = {
    0x00: 2,
    0x01: 480,
    0x02: 420,
    0x03: 360,
    0x04: 300,
    0x05: 270,
    0x06: 240,
    0x07: 210,
    0x08: 180,
    0x09: 150,
    0x0A: 120,
    0x0B: 90,
    0x0C: 60,
    0x0D: 47,
    0x0E: 43,
    0x0F: 38.5,
    0x10: 34,
    0x11: 32,
    0x12: 30,
    0x13: 28,
    0x14: 26,
    0x15: 23.5,
    0x16: 21.5,
    0x17: 19,
    0x18: 8.5,
    0x19: 6.5,
    0x1A: 4.5,
    0x1B: 2,
    0x1C: 0.5,
    0x1D: 0.3,
    0x1E: 0.2,
    0x1F: 0.1,
}



"""Security, Heath and Safety device types."""
from ..default_link import DefaultLink
from ..events import (
    ALL_CLEAR_EVENT,
    CLOSE_EVENT,
    CO_DETECTED_EVENT,
    DARK_DETECTED_EVENT,
    HEARTBEAT_EVENT,
    LEAK_DRY_EVENT,
    LEAK_WET_EVENT,
    LIGHT_DETECTED_EVENT,
    LOW_BATTERY_EVENT,
    MOTION_DETECTED_EVENT,
    MOTION_TIMEOUT_EVENT,
    NEW_DETECTED_EVENT,
    OPEN_EVENT,
    SENSOR_MALFUNCTION_EVENT,
    SMOKE_DETECTED_EVENT,
    TEST_DETECTED_EVENT,
    Event,
    HeartbeatEvent,
    LowBatteryEvent,
    WetDryEvent,
)
from ..extended_property import (
    AMBIENT_LIGHT_INTENSITY,
    BATTERY_LEVEL,
    BATTERY_LOW_LEVEL,
    HARDWARE_LIGHT_SENSITIVITY,
    HARDWARE_TIMEOUT,
    HEARBEAT_INTERVAL,
    LED_BRIGHTNESS,
    LIGHT_SENSITIVITY,
    MOTION_TIMEOUT,
    SENSOR_STATUS,
)
from ..groups import (
    CO_SENSOR,
    DOOR_SENSOR,
    HEARTBEAT,
    LEAK_SENSOR_DRY,
    LEAK_SENSOR_WET,
    LIGHT_SENSOR,
    LOW_BATTERY,
    MOTION_SENSOR,
    NEW_SENSOR,
    SENSOR_MALFUNCTION,
    SMOKE_SENSOR,
    TEST_SENSOR,
)
from ..groups.on_off import Heartbeat, LowBattery, OnOff
from ..groups.wet_dry import Wet, Dry
from ..handlers.from_device.on_level import OnLevelInbound
from ..managers.heartbeat_manager import HeartbeatManager
from ..managers.low_batter_manager import LowBatteryManager
from ..managers.on_level_manager import OnLevelManager
from ..managers.wet_dry_manager import WetDryManager
from ..operating_flag import (
    CLEANUP_REPORT_ON,
    DATABASE_DELTA,
    HARDWARE_LED_OFF,
    HARDWARE_NIGHT_MODE,
    HARDWARE_SEND_ON_ONLY,
    HEART_BEAT_ON,
    IGNORE_JUMPER_ON,
    LED_BLINK_ON_TX_ON,
    LED_ON,
    LINK_TO_FF_GROUP,
    MULTI_SEND_ON,
    NIGHT_MODE_ONLY,
    PROGRAM_LOCK_ON,
    REPEAT_CLOSED_ON,
    REPEAT_OPEN_ON,
    SEND_ON_ONLY,
    SOFTWARE_SUPPORT_ON,
    STAY_AWAKE_ON,
    TWO_GROUPS_ON,
)
from .battery_base import BatteryDeviceBase
from .device_base import Device
from .on_off_controller_base import OnOffControllerBase
from .open_close_controller_base import NormallyOpenControllerBase


class SecurityHealthSafety(Device):
    """Security, Health and Safety base device."""


class SecurityHealthSafety_DoorSensor(BatteryDeviceBase, OnOffControllerBase):
    """Door sensor model 2845-222 or similar."""

    DOOR_GROUP = 1
    LOW_BATTERY_GROUP = 3
    HEARTBEAT_GROUP = 4

    def __init__(self, address, cat, subcat, firmware=0x00, description="", model=""):
        """Init the SecurityHealthSafety_DoorSensor class."""
        buttons = {1: DOOR_SENSOR}
        super().__init__(
            address=address,
            cat=cat,
            subcat=subcat,
            firmware=firmware,
            description=description,
            model=model,
            buttons=buttons,
            on_event_name=OPEN_EVENT,
            off_event_name=CLOSE_EVENT,
        )

    def _register_handlers_and_managers(self):
        """Register handlers and managers."""
        super()._register_handlers_and_managers()
        self._low_battery_manger = LowBatteryManager(
            self._address, self.LOW_BATTERY_GROUP
        )
        self._heartbeat_manger = HeartbeatManager(self._address, self.HEARTBEAT_GROUP)

    def _register_groups(self):
        """Register groups for the Door Sensor."""
        super()._register_groups()
        self._groups[self.LOW_BATTERY_GROUP] = LowBattery(
            name=LOW_BATTERY, address=self._address, group=self.LOW_BATTERY_GROUP
        )

        self._groups[self.HEARTBEAT_GROUP] = Heartbeat(
            name=HEARTBEAT,
            address=self._address,
            group=self.HEARTBEAT_GROUP,
            default=False,
        )

    def _register_events(self):
        """Register events for the Door Sensor."""
        super()._register_events()
        self._events[LOW_BATTERY_EVENT] = LowBatteryEvent(
            name=LOW_BATTERY_EVENT, address=self._address, group=self.LOW_BATTERY_GROUP
        )
        self._events[HEARTBEAT_EVENT] = HeartbeatEvent(
            name=HEARTBEAT_EVENT, address=self._address, group=self.HEARTBEAT_GROUP
        )

    def _subscribe_to_handelers_and_managers(self):
        super()._subscribe_to_handelers_and_managers()
        self._low_battery_manger.subscribe(
            self._groups[self.LOW_BATTERY_GROUP].set_value
        )
        self._heartbeat_manger.subscribe_on(self._groups[self.DOOR_GROUP].set_value)
        self._heartbeat_manger.subscribe_off(self._groups[self.DOOR_GROUP].set_value)
        self._heartbeat_manger.subscribe(self._groups[self.HEARTBEAT_GROUP].set_value)
        self._low_battery_manger.subscribe_low_battery_event(
            self._events[LOW_BATTERY_EVENT].trigger
        )
        self._heartbeat_manger.subscribe(self._events[HEARTBEAT_EVENT].trigger)

    def _register_operating_flags(self):
        """Register operating flags for Door Sensor."""
        super()._register_operating_flags()

        self._add_operating_flag(CLEANUP_REPORT_ON, 0, 1, 16, 17)
        self._add_operating_flag(TWO_GROUPS_ON, 0, 1, 4, 5)
        self._add_operating_flag(REPEAT_OPEN_ON, 0, 2, 0x10, 0x11)
        self._add_operating_flag(REPEAT_CLOSED_ON, 0, 3, 8, 9)
        self._add_operating_flag(LINK_TO_FF_GROUP, 0, 4, 6, 7)
        self._add_operating_flag(LED_ON, 0, 5, 2, 3, is_reversed=True)
        self._add_operating_flag(PROGRAM_LOCK_ON, 0, 7, 0, 1)

        self._add_operating_flag(STAY_AWAKE_ON, 0, 6, 18, 19)

        self._add_operating_flag(DATABASE_DELTA, 1, None, None, None)

        self._add_property(name=BATTERY_LEVEL, data_field=4, set_cmd=None)
        self._add_property(name=SENSOR_STATUS, data_field=5, set_cmd=None)
        self._add_property(name=HEARBEAT_INTERVAL, data_field=6, set_cmd=2)
        self._add_property(name=BATTERY_LOW_LEVEL, data_field=6, set_cmd=3)

    def _register_default_links(self):
        super()._register_default_links()
        link_battery = DefaultLink(
            is_controller=True,
            group=self.LOW_BATTERY_GROUP,
            dev_data1=255,
            dev_data2=28,
            dev_data3=self.LOW_BATTERY_GROUP,
            modem_data1=0,
            modem_data2=0,
            modem_data3=0,
        )
        link_heartbeat = DefaultLink(
            is_controller=True,
            group=self.HEARTBEAT_GROUP,
            dev_data1=255,
            dev_data2=28,
            dev_data3=self.HEARTBEAT_GROUP,
            modem_data1=0,
            modem_data2=0,
            modem_data3=0,
        )
        self._default_links.append(link_battery)
        self._default_links.append(link_heartbeat)


class SecurityHealthSafety_OpenCloseSensor(
    BatteryDeviceBase, NormallyOpenControllerBase
):
    """Normally Open sensor."""

    def _register_operating_flags(self):
        super()._register_operating_flags()
        self._add_property(name=LED_BRIGHTNESS, data_field=3, set_cmd=0x02)
        self._add_property(name=CLEANUP_REPORT_ON, data_field=6, set_cmd=0x05, bit=0)
        self._add_property(name=IGNORE_JUMPER_ON, data_field=6, set_cmd=0x05, bit=1)
        self._add_property(name=TWO_GROUPS_ON, data_field=6, set_cmd=0x05, bit=2)
        self._add_property(name=REPEAT_OPEN_ON, data_field=6, set_cmd=0x05, bit=3)
        self._add_property(name=REPEAT_CLOSED_ON, data_field=6, set_cmd=0x05, bit=4)
        self._add_property(name=LED_ON, data_field=6, set_cmd=0x05, bit=5)
        self._add_property(name=LINK_TO_FF_GROUP, data_field=6, set_cmd=0x05, bit=6)
        self._add_property(name=PROGRAM_LOCK_ON, data_field=6, set_cmd=0x05, bit=7)


class SecurityHealthSafety_MotionSensor(BatteryDeviceBase, OnOffControllerBase):
    """Motion Sensor."""

    MOTION_GROUP = 1
    LIGHT_GROUP = 2
    LOW_BATTERY_GROUP = 3

    def __init__(self, address, cat, subcat, firmware=0x00, description="", model=""):
        """Init the SecurityHealthSafety_DoorSensor class."""
        buttons = {1: MOTION_SENSOR}
        self._light_manager = OnLevelManager(address, self.LIGHT_GROUP)
        self._low_battery_manager = LowBatteryManager(address, self.LOW_BATTERY_GROUP)
        super().__init__(
            address=address,
            cat=cat,
            subcat=subcat,
            firmware=firmware,
            description=description,
            model=model,
            buttons=buttons,
            on_event_name=MOTION_DETECTED_EVENT,
            off_event_name=MOTION_TIMEOUT_EVENT,
        )

    def _register_groups(self):
        """Register groups for the Door Sensor."""
        super()._register_groups()
        # This list state may be reversed where 0x11 means no light and 0x13 means light
        self._groups[self.LIGHT_GROUP] = OnOff(
            LIGHT_SENSOR, self._address, self.LIGHT_GROUP
        )
        self._groups[self.LOW_BATTERY_GROUP] = LowBattery(
            LOW_BATTERY, self._address, self.LOW_BATTERY_GROUP
        )

    def _register_events(self):
        """Register events for the Door Sensor."""
        super()._register_events()
        self._events[self.LIGHT_GROUP] = {}
        self._events[self.LIGHT_GROUP][LIGHT_DETECTED_EVENT] = Event(
            LIGHT_DETECTED_EVENT, self._address, self.LIGHT_GROUP
        )
        self._events[self.LIGHT_GROUP][DARK_DETECTED_EVENT] = Event(
            DARK_DETECTED_EVENT, self._address, self.LIGHT_GROUP
        )
        self._events[self.LOW_BATTERY_GROUP] = {}
        self._events[self.LOW_BATTERY_GROUP][LOW_BATTERY_EVENT] = LowBatteryEvent(
            LOW_BATTERY_EVENT, self._address, self.LOW_BATTERY_GROUP
        )

    def _subscribe_to_handelers_and_managers(self):
        super()._subscribe_to_handelers_and_managers()
        self._light_manager.subscribe(self._groups[self.LIGHT_GROUP].set_value)
        self._low_battery_manager.subscribe(
            self._groups[self.LOW_BATTERY_GROUP].set_value
        )

        self._light_manager.subscribe_on(
            self._events[self.LIGHT_GROUP][LIGHT_DETECTED_EVENT].trigger
        )
        self._light_manager.subscribe_off(
            self._events[self.LIGHT_GROUP][DARK_DETECTED_EVENT].trigger
        )
        self._low_battery_manager.subscribe_low_battery_event(
            self._events[self.LOW_BATTERY_GROUP][LOW_BATTERY_EVENT].trigger
        )

    def _register_operating_flags(self):
        super()._register_operating_flags()
        self._add_property(name=LED_BRIGHTNESS, data_field=3, set_cmd=0x02)
        self._add_property(name=MOTION_TIMEOUT, data_field=4, set_cmd=0x03)
        self._add_property(name=LIGHT_SENSITIVITY, data_field=5, set_cmd=0x04)

        self._add_property(name=SEND_ON_ONLY, data_field=6, set_cmd=0x05, bit=1)
        self._add_property(name=NIGHT_MODE_ONLY, data_field=6, set_cmd=0x05, bit=2)
        self._add_property(name=LED_ON, data_field=6, set_cmd=0x05, bit=3)
        self._add_property(name=MULTI_SEND_ON, data_field=6, set_cmd=0x05, bit=4)

        self._add_property(name=HARDWARE_TIMEOUT, data_field=7, set_cmd=None)
        self._add_property(name=HARDWARE_LIGHT_SENSITIVITY, data_field=8, set_cmd=None)

        self._add_property(name=SOFTWARE_SUPPORT_ON, data_field=9, set_cmd=None, bit=0)
        self._add_property(
            name=HARDWARE_SEND_ON_ONLY, data_field=9, set_cmd=None, bit=1
        )
        self._add_property(name=HARDWARE_NIGHT_MODE, data_field=9, set_cmd=None, bit=2)
        self._add_property(name=HARDWARE_LED_OFF, data_field=9, set_cmd=None, bit=3)

        self._add_property(name=AMBIENT_LIGHT_INTENSITY, data_field=11, set_cmd=None)
        self._add_property(name=BATTERY_LEVEL, data_field=12, set_cmd=None)

    def _register_default_links(self):
        super()._register_default_links()
        link_light = DefaultLink(
            is_controller=True,
            group=self.LIGHT_GROUP,
            dev_data1=255,
            dev_data2=28,
            dev_data3=self.LIGHT_GROUP,
            modem_data1=0,
            modem_data2=0,
            modem_data3=0,
        )
        link_battery = DefaultLink(
            is_controller=True,
            group=self.LOW_BATTERY_GROUP,
            dev_data1=255,
            dev_data2=28,
            dev_data3=self.LOW_BATTERY_GROUP,
            modem_data1=0,
            modem_data2=0,
            modem_data3=0,
        )
        self._default_links.append(link_light)
        self._default_links.append(link_battery)


class SecurityHealthSafety_LeakSensor(BatteryDeviceBase, Device):
    """Leak Sensor device."""

    DRY_GROUP = 1
    WET_GROUP = 2
    HEARTBEAT_GROUP = 4
    WET_DRY_MANGER = "manager"

    def _register_handlers_and_managers(self):
        """Regsister handlers and managers for the Leak Sensor."""
        super()._register_handlers_and_managers()
        self._managers[self.WET_DRY_MANGER] = WetDryManager(self._address)
        self._managers[self.HEARTBEAT_GROUP] = HeartbeatManager(
            self._address, self.HEARTBEAT_GROUP
        )

    def _register_groups(self):
        """Register groups for the Door Sensor."""
        # Group 1 is the Dry state. This uses the following definition:
        #   Dry = Closed
        #   Wet = Open
        # When an ON message (0x11) is received for group 1 the sensor is dry.
        # When an ON message (0x11) is received for group 2 the sensor is wet.
        # Dry state is handled via the Dry event below

        self._groups[self.DRY_GROUP] = Dry(
            LEAK_SENSOR_DRY, self._address, self.DRY_GROUP
        )
        self._groups[self.WET_GROUP] = Wet(
            LEAK_SENSOR_WET, self._address, self.WET_GROUP
        )
        self._groups[self.HEARTBEAT_GROUP] = Heartbeat(
            HEARTBEAT, self._address, self.HEARTBEAT_GROUP, default=False
        )

    def _register_events(self):
        """Register events for the Door Sensor."""
        self._events[LEAK_DRY_EVENT] = WetDryEvent(
            LEAK_DRY_EVENT, self._address, self.DRY_GROUP
        )
        self._events[LEAK_WET_EVENT] = WetDryEvent(
            LEAK_WET_EVENT, self._address, self.WET_GROUP
        )
        self._events[HEARTBEAT_EVENT] = HeartbeatEvent(
            HEARTBEAT_EVENT, self._address, self.HEARTBEAT_GROUP
        )

    def _subscribe_to_handelers_and_managers(self):
        """Subscribe to handlers and managers."""
        super()._subscribe_to_handelers_and_managers()
        dry_group = self._groups[self.DRY_GROUP]
        wet_group = self._groups[self.WET_GROUP]
        dry_event = self._events[LEAK_DRY_EVENT]
        wet_event = self._events[LEAK_WET_EVENT]
        hb_event = self._events[HEARTBEAT_EVENT]
        wet_dry_mgr = self._managers[self.WET_DRY_MANGER]
        hb_mgr = self._managers[self.HEARTBEAT_GROUP]

        wet_dry_mgr.subscribe(dry_group.set_value)
        wet_dry_mgr.subscribe(wet_group.set_value)

        wet_dry_mgr.subscribe_dry(dry_event.trigger)
        wet_dry_mgr.subscribe_wet(wet_event.trigger)

        hb_mgr.subscribe(hb_event.trigger)
        hb_mgr.subscribe_on(self._heartbeat_dry)
        hb_mgr.subscribe_off(self._heartbeat_wet)

    def _heartbeat_dry(self, on_level):
        """Receive heartbeat on/off message and create wet/dry status."""
        self._groups[self.DRY_GROUP].value = True
        self._groups[self.WET_GROUP].value = False
        self._events[LEAK_DRY_EVENT].trigger(on_level)

    def _heartbeat_wet(self, on_level):
        """Receive heartbeat on/off message and create wet/dry status."""
        self._groups[self.DRY_GROUP].value = False
        self._groups[self.WET_GROUP].value = True
        self._events[LEAK_WET_EVENT].trigger(on_level)

    def _register_operating_flags(self):
        # bit 0 = Cleanup Report
        # bit 1 = Don’t Read the Jumper
        # bit 2 = 2 Groups
        # bit 3 = Repeat Open
        # bit 4 = Repeat Closed
        # bit 5 = LED On/Off
        # bit 6 = Link to FF Group
        # bit 7 = Programming Lock
        # (0x4F is the default Config Byte)

        self._add_property(name=LED_BRIGHTNESS, data_field=3, set_cmd=2)

        self._add_property(name=CLEANUP_REPORT_ON, data_field=6, set_cmd=5, bit=0)
        self._add_property(name=IGNORE_JUMPER_ON, data_field=6, set_cmd=5, bit=1)
        self._add_property(name=TWO_GROUPS_ON, data_field=6, set_cmd=5, bit=2)
        self._add_property(name=REPEAT_OPEN_ON, data_field=6, set_cmd=5, bit=3)
        self._add_property(name=REPEAT_CLOSED_ON, data_field=6, set_cmd=5, bit=4)
        self._add_property(name=LED_ON, data_field=6, set_cmd=5, bit=5)
        self._add_property(name=LINK_TO_FF_GROUP, data_field=6, set_cmd=5, bit=6)
        self._add_property(name=PROGRAM_LOCK_ON, data_field=6, set_cmd=5, bit=7)

    def _register_default_links(self):
        super()._register_default_links()
        link_dry = DefaultLink(
            is_controller=True,
            group=self.DRY_GROUP,
            dev_data1=255,
            dev_data2=28,
            dev_data3=self.DRY_GROUP,
            modem_data1=0,
            modem_data2=0,
            modem_data3=0,
        )
        link_wet = DefaultLink(
            is_controller=True,
            group=self.WET_GROUP,
            dev_data1=255,
            dev_data2=28,
            dev_data3=self.WET_GROUP,
            modem_data1=0,
            modem_data2=0,
            modem_data3=0,
        )
        link_heartbeat = DefaultLink(
            is_controller=True,
            group=self.HEARTBEAT_GROUP,
            dev_data1=255,
            dev_data2=28,
            dev_data3=self.HEARTBEAT_GROUP,
            modem_data1=0,
            modem_data2=0,
            modem_data3=0,
        )
        self._default_links.append(link_dry)
        self._default_links.append(link_wet)
        self._default_links.append(link_heartbeat)


class SecurityHealthSafety_Smokebridge(Device):
    """Smokebridge device."""

    SMOKE_DETECTED_GROUP = 1
    CO_DETECTED_GROUP = 2
    TEST_DETECTED_GROUP = 3
    NEW_DETECTED_GROUP = 4
    ALL_CLEAR_GROUP = 5
    LOW_BATTERY_GROUP = 6
    SENSOR_MALFUNCTION_GROUP = 7
    HEARTBEAT_GROUP = 0x0A

    def _register_handlers_and_managers(self):
        super()._register_handlers_and_managers()
        self._handlers[self.SMOKE_DETECTED_GROUP] = OnLevelInbound(
            self._address, self.SMOKE_DETECTED_GROUP
        )
        self._handlers[self.CO_DETECTED_GROUP] = OnLevelInbound(
            self._address, self.CO_DETECTED_GROUP
        )
        self._handlers[self.TEST_DETECTED_GROUP] = OnLevelInbound(
            self._address, self.TEST_DETECTED_GROUP
        )
        self._handlers[self.NEW_DETECTED_GROUP] = OnLevelInbound(
            self._address, self.NEW_DETECTED_GROUP
        )
        self._handlers[self.SENSOR_MALFUNCTION_GROUP] = OnLevelInbound(
            self._address, self.SENSOR_MALFUNCTION_GROUP
        )

        self._handlers[self.ALL_CLEAR_GROUP] = OnLevelInbound(
            self._address, self.ALL_CLEAR_GROUP
        )
        self._managers[self.LOW_BATTERY_GROUP] = LowBatteryManager(
            self._address, self.LOW_BATTERY_GROUP
        )
        self._managers[self.HEARTBEAT_GROUP] = HeartbeatManager(
            self._address, self.HEARTBEAT_GROUP
        )

    def _register_groups(self):
        super()._register_groups()
        self._groups[self.SMOKE_DETECTED_GROUP] = OnOff(
            SMOKE_SENSOR, self._address, self.SMOKE_DETECTED_GROUP
        )
        self._groups[self.CO_DETECTED_GROUP] = OnOff(
            CO_SENSOR, self._address, self.CO_DETECTED_GROUP
        )
        self._groups[self.TEST_DETECTED_GROUP] = OnOff(
            TEST_SENSOR, self._address, self.TEST_DETECTED_GROUP
        )
        self._groups[self.NEW_DETECTED_GROUP] = OnOff(
            NEW_SENSOR, self._address, self.NEW_DETECTED_GROUP
        )
        self._groups[self.LOW_BATTERY_GROUP] = LowBattery(
            LOW_BATTERY, self._address, self.LOW_BATTERY_GROUP
        )
        self._groups[self.SENSOR_MALFUNCTION_GROUP] = OnOff(
            SENSOR_MALFUNCTION, self._address, self.SENSOR_MALFUNCTION_GROUP
        )
        self._groups[self.HEARTBEAT_GROUP] = Heartbeat(
            HEARTBEAT, self._address, self.HEARTBEAT_GROUP, default=False
        )

    def _register_events(self):
        super()._register_events()
        self._events[SMOKE_DETECTED_EVENT] = Event(
            SMOKE_DETECTED_EVENT, self._address, self.SMOKE_DETECTED_GROUP
        )
        self._events[CO_DETECTED_EVENT] = Event(
            CO_DETECTED_EVENT, self._address, self.CO_DETECTED_GROUP
        )
        self._events[TEST_DETECTED_EVENT] = Event(
            TEST_DETECTED_EVENT, self._address, self.TEST_DETECTED_GROUP
        )
        self._events[NEW_DETECTED_EVENT] = Event(
            NEW_DETECTED_EVENT, self._address, self.NEW_DETECTED_GROUP
        )
        self._events[ALL_CLEAR_EVENT] = Event(
            ALL_CLEAR_EVENT, self._address, self.ALL_CLEAR_GROUP
        )
        self._events[LOW_BATTERY_EVENT] = LowBatteryEvent(
            LOW_BATTERY_EVENT, self._address, self.LOW_BATTERY_GROUP
        )
        self._events[SENSOR_MALFUNCTION_EVENT] = Event(
            SENSOR_MALFUNCTION_EVENT, self._address, self.SENSOR_MALFUNCTION_GROUP
        )
        self._events[HEARTBEAT_EVENT] = HeartbeatEvent(
            HEARTBEAT_EVENT, self._address, self.HEARTBEAT_GROUP
        )

    def _subscribe_to_handelers_and_managers(self):
        super()._subscribe_to_handelers_and_managers()

        self._handlers[self.SMOKE_DETECTED_GROUP].subscribe(
            self._groups[self.SMOKE_DETECTED_GROUP].set_value
        )
        self._handlers[self.SMOKE_DETECTED_GROUP].subscribe(
            self._events[SMOKE_DETECTED_EVENT].trigger
        )

        self._handlers[self.CO_DETECTED_GROUP].subscribe(
            self._groups[self.CO_DETECTED_GROUP].set_value
        )
        self._handlers[self.CO_DETECTED_GROUP].subscribe(
            self._events[CO_DETECTED_EVENT].trigger
        )

        self._handlers[self.TEST_DETECTED_GROUP].subscribe(
            self._groups[self.TEST_DETECTED_GROUP].set_value
        )
        self._handlers[self.TEST_DETECTED_GROUP].subscribe(
            self._events[TEST_DETECTED_EVENT].trigger
        )

        self._handlers[self.NEW_DETECTED_GROUP].subscribe(
            self._groups[self.NEW_DETECTED_GROUP].set_value
        )
        self._handlers[self.NEW_DETECTED_GROUP].subscribe(
            self._events[NEW_DETECTED_EVENT].trigger
        )

        self._handlers[self.SENSOR_MALFUNCTION_GROUP].subscribe(
            self._groups[self.SENSOR_MALFUNCTION_GROUP].set_value
        )
        self._handlers[self.SENSOR_MALFUNCTION_GROUP].subscribe(
            self._events[SENSOR_MALFUNCTION_EVENT].trigger
        )

        self._handlers[self.ALL_CLEAR_GROUP].subscribe(self._all_clear_received)

    def _all_clear_received(self, on_level):
        """All-Clear message received."""
        if self._groups[self.SMOKE_DETECTED_GROUP].value:
            self._groups[self.SMOKE_DETECTED_GROUP].set_value(0)
        if self._groups[self.CO_DETECTED_GROUP].value:
            self._groups[self.CO_DETECTED_GROUP].set_value(0)
        if self._groups[self.TEST_DETECTED_GROUP].value:
            self._groups[self.TEST_DETECTED_GROUP].set_value(0)
        if self._groups[self.NEW_DETECTED_GROUP].value:
            self._groups[self.NEW_DETECTED_GROUP].set_value(0)
        if self._groups[self.SENSOR_MALFUNCTION_GROUP].value:
            self._groups[self.SENSOR_MALFUNCTION_GROUP].set_value(0)

    def _register_operating_flags(self):
        super()._register_operating_flags()
        self._add_operating_flag(
            name=PROGRAM_LOCK_ON, group=0, bit=0, set_cmd=0, unset_cmd=1
        )
        self._add_operating_flag(
            name=LED_BLINK_ON_TX_ON, group=0, bit=1, set_cmd=2, unset_cmd=3
        )
        self._add_operating_flag(
            name=LED_ON, group=0, bit=4, set_cmd=8, unset_cmd=9, is_reversed=True
        )
        self._add_operating_flag(
            name=HEART_BEAT_ON, group=0, bit=5, set_cmd=6, unset_cmd=7
        )
        self._add_operating_flag(
            name=CLEANUP_REPORT_ON, group=0, bit=6, set_cmd=0x0B, unset_cmd=0x0A
        )

    def _register_default_links(self):
        super()._register_default_links()

        link_smoke = DefaultLink(
            is_controller=True,
            group=self.SMOKE_DETECTED_GROUP,
            dev_data1=255,
            dev_data2=28,
            dev_data3=self.SMOKE_DETECTED_GROUP,
            modem_data1=0,
            modem_data2=0,
            modem_data3=0,
        )
        link_co = DefaultLink(
            is_controller=True,
            group=self.CO_DETECTED_GROUP,
            dev_data1=255,
            dev_data2=28,
            dev_data3=self.CO_DETECTED_GROUP,
            modem_data1=0,
            modem_data2=0,
            modem_data3=0,
        )
        link_test = DefaultLink(
            is_controller=True,
            group=self.TEST_DETECTED_GROUP,
            dev_data1=255,
            dev_data2=28,
            dev_data3=self.TEST_DETECTED_GROUP,
            modem_data1=0,
            modem_data2=0,
            modem_data3=0,
        )
        link_new = DefaultLink(
            is_controller=True,
            group=self.NEW_DETECTED_GROUP,
            dev_data1=255,
            dev_data2=28,
            dev_data3=self.NEW_DETECTED_GROUP,
            modem_data1=0,
            modem_data2=0,
            modem_data3=0,
        )
        link_clear = DefaultLink(
            is_controller=True,
            group=self.ALL_CLEAR_GROUP,
            dev_data1=255,
            dev_data2=28,
            dev_data3=self.ALL_CLEAR_GROUP,
            modem_data1=0,
            modem_data2=0,
            modem_data3=0,
        )
        link_battery = DefaultLink(
            is_controller=True,
            group=self.LOW_BATTERY_GROUP,
            dev_data1=255,
            dev_data2=28,
            dev_data3=self.LOW_BATTERY_GROUP,
            modem_data1=0,
            modem_data2=0,
            modem_data3=0,
        )
        link_malfunction = DefaultLink(
            is_controller=True,
            group=self.SENSOR_MALFUNCTION_GROUP,
            dev_data1=255,
            dev_data2=28,
            dev_data3=self.SENSOR_MALFUNCTION_GROUP,
            modem_data1=0,
            modem_data2=0,
            modem_data3=0,
        )
        link_heartbeat = DefaultLink(
            is_controller=True,
            group=self.HEARTBEAT_GROUP,
            dev_data1=255,
            dev_data2=28,
            dev_data3=self.HEARTBEAT_GROUP,
            modem_data1=0,
            modem_data2=0,
            modem_data3=0,
        )

        self._default_links.append(link_smoke)
        self._default_links.append(link_co)
        self._default_links.append(link_test)
        self._default_links.append(link_new)
        self._default_links.append(link_clear)
        self._default_links.append(link_battery)
        self._default_links.append(link_malfunction)
        self._default_links.append(link_heartbeat)



"""Operating flags list."""
from .address import Address
from .device_flag import DeviceFlagBase

PROGRAM_LOCK_ON = "program_lock_on"
BUTTON_LOCK_ON = "button_lock_on"
LED_ON = "led_on"
LED_OFF = "led_off"
KEY_BEEP_ON = "key_beep_on"
STAY_AWAKE_ON = "stay_awake_on"
LISTEN_ONLY_ON = "listen_only_on"
HEART_BEAT_ON = "heart_beat_on"
HEART_BEAT_OFF = "heart_beat_off"
SEND_ON_ONLY = "send_on_only_on"
GROUPED_ON = "grouped_on"
TWO_GROUPS_ON = "two_groups_on"
RESUME_DIM_ON = "resume_dim_on"
RF_DISABLE_ON = "rf_disable_on"
POWERLINE_DISABLE_ON = "powerline_disable_on"
TEND_ON = "tend_on"
X10_OFF = "x10_off"
CLEANUP_REPORT_ON = "cleanup_report_on"
CLEANUP_REPORT_OFF = "cleanup_report_off"
DETACH_LOAD_ON = "detach_load_on"
SMART_HOPS_ON = "smart_hops_on"
LED_BLINK_ON_ERROR_ON = "blink_on_error_on"
LED_BLINK_ON_ERROR_OFF = "blink_on_error_off"
LED_BLINK_ON_TX_ON = "blink_on_tx_on"
REPEAT_OPEN_ON = "repeat_open_on"
REPEAT_CLOSED_ON = "repeat_closed_on"
LINK_TO_FF_GROUP = "link_to_ff_group"
LOAD_SENSE_ON = "load_sense_on"
LOAD_SENSE_2_ON = "load_sense_2_on"
IGNORE_JUMPER_ON = "ignore_jumper_on"
NIGHT_MODE_ONLY = "night_mode_only"
MULTI_SEND_ON = "multi_send_on"
SOFTWARE_SUPPORT_ON = "software_support_on"
HARDWARE_SEND_ON_ONLY = "hardware_send_on_only"
HARDWARE_NIGHT_MODE = "hardware_night_mode"
HARDWARE_LED_OFF = "hardware_led_off"
RELAY_ON_SENSE_ON = "relay_on_sense_on"
MOMENTARY_MODE_ON = "momentary_on"
MOMENTARY_ON_OFF_TRIGGER = "momentary_on_off_trigger"
SENSE_SENDS_OFF = "sense_sends_off"
MOMENTARY_FOLLOW_SENSE = "momentary_follow_sense"
INSTEON_OFF = "insteon_off"
CHECKSUM_OFF = "checksum_off"
STANDARD_HOLDOFF = "standard_holdoff"


DUAL_LINE_ON = "dual_line_on"
LATCHING_ON = "latching_on"
THREE_WAY_ON = "three_way_on"
REVERSED_ON = "reversed_on"
FORWARD_ON = "forward_on"
MOMENTARY_LINE_ON = "momentary_line_on"
NOT_3_WAY = "not_3_way"

CRC_ERROR_COUNT = "crc_error_count"
SIGNAL_TO_NOISE_OF_LAST_FAILURE = "signal_to_noise_of_last_failure"
SIGNAL_TO_NOISE_FAILURE_COUNT = "signal_to_noise_failure_count"
DATABASE_DELTA = "database_delta"

CELSIUS = "celsius"
TIME_24_HOUR_FORMAT = "time_24_hour_format"
SOFTWARE_LOCK_ON = "SOFTWARE_LOCK_ON"


class OperatingFlag(DeviceFlagBase):
    """Operating flag for a device."""

    def __init__(self, address, name, flag_type: type, is_reversed=False):
        """Init the OperatingFlag class."""
        self._address = Address(address)
        topic = "{}.operating_flag.{}".format(self._address.id, name)
        super().__init__(topic, name, flag_type, is_reversed)



"""Manage writes to the All-Link Database."""
import asyncio

from ..aldb.aldb_record import ALDBRecord
from ..handlers.to_device.write_aldb import WriteALDBCommandHandler


class ALDBWriteException(Exception):
    """All-Link Database write excetion."""


class ALDBWriteManager:
    """Manage writes to the All-Link Database."""

    def __init__(self, aldb):
        """Init the ALDBWriteManager class."""
        self._aldb = aldb
        self._address = aldb.address
        self._write_handler = WriteALDBCommandHandler(self._address)

    @property
    def can_write(self):
        """Return if the ALDB is loaded and we can write to it."""
        return self._aldb.is_loaded

    def write(self, record: ALDBRecord):
        """Write the ALDB record."""
        asyncio.ensure_future(self.async_write(record))

    async def async_write(self, record: ALDBRecord):
        """Write the ALDB record."""
        if not self.can_write:
            raise ALDBWriteException(
                "All-Link databased must be loaded before writing."
            )

        if record.mem_addr > self._aldb.first_mem_addr:
            raise ALDBWriteException("All-link record memory address out of range.")

        # If we get a direct ACK, we assume the ALDB has been written.
        # Should we force read to ensure it is accuate.
        return await self._write_handler.async_send(
            mem_addr=record.mem_addr,
            controller=record.is_controller,
            group=record.group,
            target=record.target,
            data1=record.data1,
            data2=record.data2,
            data3=record.data3,
            in_use=record.is_in_use,
            bit5=record.is_bit5_set,
            bit4=record.is_bit4_set,
        )



"""Manage identifying unknown devices."""
import asyncio
import logging
from collections import namedtuple

from binascii import unhexlify

from .. import pub
from ..utils import subscribe_topic, unsubscribe_topic
from ..address import Address
from ..handlers.from_device.assign_to_all_link_group import AssignToAllLinkGroupCommand
from ..handlers.all_link_completed import AllLinkCompletedHandler
from ..handlers.from_device.delete_from_all_link_group import (
    DeleteFromAllLinkGroupCommand,
)
from ..handlers.to_device.id_request import IdRequestCommand
from ..subscriber_base import SubscriberBase

_LOGGER = logging.getLogger(__name__)
MAX_RETRIES = 5
RETRY_PAUSE = 2


DeviceId = namedtuple("DeviceId", "address cat subcat firmware")  # product_id')


def _normalize_identifier(value):
    """Convert a byte, str or int to int."""
    if value is None:
        return value
    if isinstance(value, (bytearray, bytes)):
        return int.from_bytes(value, "big")
    if isinstance(value, str):
        if value[0:2] == "0x":
            value = value[2:]
        return int.from_bytes(unhexlify(value), "big")
    return int(value)


class DeviceIdManager(SubscriberBase):
    """Manage device identities."""

    def __init__(self):
        """Init the DeviceIdManager class."""
        super().__init__(subscriber_topic="device_id")
        self._unknown_devices = []
        self._device_ids = {}
        self._awake_devices = []
        # Cannot set queue here because we are outside the loop
        self._awake_devices_queue = None
        self._id_device_lock = asyncio.Lock()
        self._all_link_complete = AllLinkCompletedHandler()
        self._all_link_complete.subscribe(self._all_link_complete_received)

    def __getitem__(self, address):
        """Return the unknown device list."""
        address = Address(address)
        return self._device_ids.get(address)

    def start(self):
        """Start the ID manager for unknown devices."""
        asyncio.ensure_future(self._id_awake_devices())

    def close(self):
        """Close the ID listener."""
        if self._awake_devices_queue is not None:
            self._awake_devices_queue.put_nowait(None)

    def append(self, address: Address):
        """Append a device address to the list."""
        address = Address(address)
        # If the address is already in the unknown device list stop
        if address in self._unknown_devices:
            return

        # If it is not in the identified device list, add it as an unknown device
        if self._device_ids.get(address) is None:
            self._device_ids[address] = DeviceId(None, None, None, None)
            self._unknown_devices.append(address)
            subscribe_topic(self._device_awake, address.id)

    def set_device_id(
        self, address: Address, cat: int, subcat: int, firmware: int = 0x00
    ):
        """Set the device ID of a device."""
        address = Address(address)
        cat = _normalize_identifier(cat)
        subcat = _normalize_identifier(subcat)
        firmware = _normalize_identifier(firmware)
        try:
            self._unknown_devices.remove(address)
        except ValueError:
            pass
        device_id = DeviceId(address, cat, subcat, firmware)
        self._device_ids[address] = device_id
        self._call_subscribers(device_id=device_id)

    async def async_id_devices(self, refresh: bool = False):
        """Identify the devices in the unknown device list."""
        if refresh:
            for address_id in self._device_ids:
                self._device_ids[address_id] = DeviceId(None, None, None, None)
                address = Address(address_id)
                if address not in self._unknown_devices:
                    self._unknown_devices.append(address)
        # We change the unknown device list during the process so we make a copy
        address_list = self._unknown_devices.copy()
        for address in address_list:
            await self.async_id_device(address)
        return self._device_ids

    async def async_id_device(self, address: Address, refresh: bool = False):
        """Call ID Request command for all unknown devices."""
        if self._id_device_lock.locked():
            self._id_device_lock.release()
        await self._id_device_lock.acquire()
        address = Address(address)
        self.append(address)
        if refresh:
            self._device_ids[address] = DeviceId(None, None, None, None)
        id_handler = IdRequestCommand(address)
        id_response_handler = AssignToAllLinkGroupCommand(address)
        id_response_handler.subscribe(self._id_response)
        id_response_handler_alt = DeleteFromAllLinkGroupCommand(address)
        id_response_handler_alt.subscribe(self._id_response)
        retries = 0
        while self._device_ids[address].cat is None and retries <= MAX_RETRIES:
            # TODO check for success or failure
            await id_handler.async_send()
            retries += 1
            await asyncio.sleep(RETRY_PAUSE)
        id_response_handler.unsubscribe(self._id_response)
        id_response_handler_alt.unsubscribe(self._id_response)
        if self._id_device_lock.locked():
            self._id_device_lock.release()
        return self._device_ids[address]

    def _id_response(self, address, cat, subcat, firmware, group, mode):
        address = Address(address)
        device_id = DeviceId(address, cat, subcat, firmware)
        self._device_ids[address] = device_id
        try:
            self._unknown_devices.remove(address)
        except ValueError:
            pass
        unsubscribe_topic(self._device_awake, address.id)
        self._call_subscribers(device_id=device_id)

    def _set_device_info(
        self, address, cat=None, subcat=None, firmware=None, product_id=None
    ):
        """Set the device ID info."""
        device_id = self._device_ids.get(address)
        if device_id is None:
            new_id = DeviceId(address, cat, subcat, firmware)
        else:
            new_id = DeviceId(
                address,
                cat if cat else device_id.cat,
                subcat if subcat else device_id.subcat,
                firmware if firmware else device_id.firmware,
            )
        self._device_ids[address] = new_id

    async def _id_awake_devices(self):
        """Loop on devices that wake up and send a message."""
        if self._awake_devices_queue is None:
            self._awake_devices_queue = asyncio.Queue()
        while True:
            address = await self._awake_devices_queue.get()
            if address is None:
                break
            await asyncio.sleep(0.5)
            await self.async_id_device(address=address)
            await asyncio.sleep(2)
            self._check_awake_device(address)

    def _check_awake_device(self, address):
        """Check if an awake device has been identified."""
        dev_id = self._device_ids.get(address)
        if dev_id is None or dev_id.cat is None:
            subscribe_topic(self._device_awake, address.id)
        try:
            self._awake_devices.remove(address)
        except ValueError:
            pass

    def _device_awake(self, topic=pub.AUTO_TOPIC, **kwargs):
        """Identify an unknown device.

        An unknown device has sent a message so we try to identify it.

        """
        if self._id_device_lock.locked():
            return
        try:
            address = Address(topic.name.split(".")[0])
        except ValueError:
            pass
        else:
            if address in self._awake_devices:
                return
            unsubscribe_topic(self._device_awake, address.id)
            self._awake_devices.append(address)
            if self._awake_devices_queue is None:
                self._awake_devices_queue = asyncio.Queue()
            self._awake_devices_queue.put_nowait(address)

    def _all_link_complete_received(self, mode, group, target, cat, subcat, firmware):
        """Receive All-Link complete message."""
        self._id_response(target, cat, subcat, firmware, group, mode)



"""Manage creation of default links for a device."""

import logging

from ...constants import ResponseStatus
from ...utils import multiple_status
from . import async_link_devices

TIMEOUT = 3
_LOGGER = logging.getLogger(__name__)


async def async_add_default_links(device):
    """Establish default links between the modem and device."""
    from ... import devices

    if not device.aldb.is_loaded:
        await device.aldb.async_load()

    if not device.aldb.is_loaded:
        return ResponseStatus.UNSENT

    results = []
    for link_info in device.default_links:
        is_controller = link_info.is_controller
        group = link_info.group
        if is_controller:
            controller = device
            responder = devices.modem
            data1 = link_info.modem_data1
            data2 = link_info.modem_data2
            data3 = link_info.modem_data3
        else:
            controller = devices.modem
            responder = device
            data1 = link_info.dev_data1
            data2 = link_info.dev_data2
            data3 = link_info.dev_data3
        result = await async_link_devices(
            controller, responder, group, data1, data2, data3
        )
        results.append(result)
    return multiple_status(*results)



"""All-Link Record Flags."""

from binascii import hexlify

from ...constants import AllLinkMode
from ...utils import bit_is_set, set_bit, test_values_eq


def create(
    in_use: bool,
    controller: bool,
    hwm: bool,
    bit5: bool = False,
    bit4: bool = False,
    bit3: bool = False,
    bit2: bool = False,
    bit0: bool = False,
):
    """Create an AllLinkRecordFlags entity."""
    flags = 0x00
    flags = set_bit(flags, 7, in_use)
    flags = set_bit(flags, 6, controller)
    flags = set_bit(flags, 5, bit5)
    flags = set_bit(flags, 4, bit4)
    flags = set_bit(flags, 3, bit3)
    flags = set_bit(flags, 2, bit2)
    flags = set_bit(flags, 1, not hwm)
    flags = set_bit(flags, 0, bit0)
    return AllLinkRecordFlags(flags)


def _normalize(data):
    if isinstance(data, AllLinkRecordFlags):
        return bytes(data)
    return data


class AllLinkRecordFlags:
    """All-Link Record Flags."""

    def __init__(self, data: int):
        """Init the AllLinkRecordFlags class."""
        data = _normalize(data)
        if isinstance(data, bytes):
            data = int.from_bytes(data, byteorder="big")
        self._in_use = bit_is_set(data, 7)
        is_controller = bit_is_set(data, 6)
        self._mode = AllLinkMode(0)
        if is_controller:
            self._mode = AllLinkMode(1)
        self._bit5 = bit_is_set(data, 5)
        self._bit4 = bit_is_set(data, 4)
        self._bit3 = bit_is_set(data, 3)
        self._bit2 = bit_is_set(data, 2)
        self._hwm = not bit_is_set(data, 1)
        self._bit0 = bit_is_set(data, 0)

    def __bytes__(self):
        """Return the byte representation of the flags."""
        flags = 0x00
        flags = set_bit(flags, 7, self._in_use)
        flags = set_bit(flags, 6, bool(self._mode.value))
        flags = set_bit(flags, 5, self._bit5)
        flags = set_bit(flags, 4, self._bit4)
        flags = set_bit(flags, 3, self._bit3)
        flags = set_bit(flags, 2, self._bit2)
        flags = set_bit(flags, 1, not self._hwm)
        flags = set_bit(flags, 0, self._bit0)
        return bytes([flags])

    def __int__(self):
        """Return the integer representation of the flags."""
        return int.from_bytes(bytes(self), byteorder="big")

    def __repr__(self):
        """Return the hex representation of the flags."""
        val = {
            "in use": 1 if self.is_in_use else 0,
            "mode": 1 if bool(self.mode.value) else 0,
            "bit5": 1 if self.is_bit_5_set else 0,
            "bit4": 1 if self.is_bit_4_set else 0,
            "bit3": 1 if self.is_bit_3_set else 0,
            "bit2": 1 if self.is_bit_2_set else 0,
            "hwm": 0 if self.is_hwm else 1,
            "bit0": 1 if self.is_bit_0_set else 0,
        }
        return str(val)

    def __str__(self):
        """Return the hex representation of the flags."""
        return hexlify(bytes(self)).decode()

    def __eq__(self, other):
        """Check equality of this vs other."""
        if not isinstance(other, AllLinkRecordFlags):
            return False
        match = True
        match = match & test_values_eq(self.is_bit_0_set, other.is_bit_0_set)
        match = match & test_values_eq(self.is_bit_2_set, other.is_bit_2_set)
        match = match & test_values_eq(self.is_bit_3_set, other.is_bit_3_set)
        match = match & test_values_eq(self.is_bit_4_set, other.is_bit_4_set)
        match = match & test_values_eq(self.is_bit_5_set, other.is_bit_5_set)
        match = match & test_values_eq(self.is_in_use, other.is_in_use)
        match = match & test_values_eq(self.mode, other.mode)
        match = match & test_values_eq(self.is_hwm, other.is_hwm)
        return match

    def __hash__(self):
        """Represent the AllLinkrecordFlags object as a hash."""
        return hash(bytes(self))

    @property
    def is_in_use(self):
        """Set the record in use value."""
        return self._in_use

    @property
    def mode(self):
        """Return if the record is a responder or controller."""
        return self._mode

    @property
    def is_hwm(self):
        """Return if the record is the high water mark."""
        return self._hwm

    @property
    def is_bit_5_set(self):
        """Return if bit 5 is set."""
        return self._bit5

    @property
    def is_bit_4_set(self):
        """Return if bit 4 is set."""
        return self._bit4

    @property
    def is_bit_3_set(self):
        """Return if bit 3 is set."""
        return self._bit3

    @property
    def is_bit_2_set(self):
        """Return if bit 2 is set."""
        return self._bit2

    @property
    def is_bit_0_set(self):
        """Return if bit 0 is set."""
        return self._bit0



"""Manage Insteon Scenes."""
import asyncio
import logging

from .. import devices
from ..device_types.plm import PLM
from ..handlers.send_all_link_off import SendAllLinkOffCommandHandler
from ..handlers.send_all_link_on import SendAllLinkOnCommandHandler
from .link_manager import async_link_devices

_LOGGER = logging.getLogger(__name__)


class Scenes:
    """Manage a list of scenes."""

    def __init__(self):
        """Init the Scenes class."""
        self._scenes = {}

    def __setitem__(self, group, device):
        """Add a device to a scene."""
        device_list = self._scenes.get(group, [])
        if device not in device_list:
            device_list.append(device)
            self._scenes[group] = device_list

    def __getitem__(self, group):
        """Get a list of links associated with a scene."""
        return self._scenes.get(group, [])

    def get_devices(self, group):
        """Return the devices associated with a scene (group controller tuple)."""
        scene = self._scenes.get(group)
        if scene:
            for device in scene:
                if device:
                    yield device


def identify_scenes():
    """Look through device All-Link databases and find scenes."""
    modem = devices.modem
    scenes = Scenes()
    for address in devices:
        device = devices[address]
        for mem_addr in device.aldb:
            rec = device.aldb[mem_addr]
            if (
                rec.group != 0
                and rec.is_responder
                and rec.target == modem.address
                and rec.is_in_use
            ):
                scenes[rec.group] = device
    for mem_addr in modem.aldb:
        rec = modem.aldb[mem_addr]
        if rec.is_controller and rec.group != 0:
            scenes[rec.group] = devices[rec.target]
    return scenes


async def async_add_device_to_scene(
    device, group: int, on_level=0xFF, ramp_rate=0x28, button=0
):
    """Create a new scene.

    Parameters:
    - group: Scene number.
    - device_scene_info: Iterable colleciton of SceneInfo objects.

    """
    if isinstance(devices.modem, PLM):
        await _plm_add_device_to_scene(group, device, on_level, ramp_rate, button)
    else:
        pass


async def async_trigger_scene_on(group):
    """Trigger an Insteon scene."""
    scenes = identify_scenes()

    await SendAllLinkOnCommandHandler().async_send(group=group)

    await asyncio.sleep(2)
    for device in scenes.get_devices(group):
        await device.async_status()


async def async_trigger_scene_off(group):
    """Trigger an Insteon scene."""
    scenes = identify_scenes()

    await SendAllLinkOffCommandHandler().async_send(group=group)

    await asyncio.sleep(2)
    for device in scenes.get_devices(group):
        await device.async_status()


async def _plm_add_device_to_scene(group, device, on_level, ramp_rate, button):
    modem = devices.modem
    device.aldb.add(
        group=group,
        target=modem.address,
        controller=False,
        data1=on_level,
        data2=ramp_rate,
        data3=button,
    )
    modem.aldb.add(
        group=group, target=device.address, controller=True, data1=0, data2=0, data3=0
    )
    # TODO check for success or failure
    await device.aldb.async_write()
    await modem.aldb.async_write()


async def _hub_add_device_to_scene(group, device, on_level, ramp_rate, button):
    # TODO check for success or failure
    await device.async_status()
    curr_state = device.groups[group].value
    await _set_device_state(device, on_level, button)
    await async_link_devices(devices.modem, device, group)
    await _set_device_state(device, curr_state, button)


async def _set_device_state(device, on_level, button):
    if device.cat == 0x01:
        device.on(on_level=on_level, group=button)
    elif device.cat == 0x02:
        if on_level:
            device.on(group=button)
        else:
            device.off(group=button)



"""Insteon device address class."""
import binascii
import logging

_LOGGER = logging.getLogger(__name__)


def _normalize(addr):
    """Take any format of address and turn it into a hex string."""
    normalize = None
    if isinstance(addr, Address):
        normalize = bytes(addr)

    elif isinstance(addr, bytearray):
        normalize = bytes(addr)

    elif isinstance(addr, bytes):
        normalize = addr

    elif isinstance(addr, str):
        addr_clean = addr.replace(".", "")
        try:
            if len(addr_clean) != 6:
                raise ValueError("Improper address value: {}".format(addr))
            normalize = binascii.unhexlify(addr_clean.lower())
        except binascii.Error:
            raise ValueError("Improper address value: {}".format(addr))
    return normalize


class Address:
    """Datatype definition for INSTEON device address handling."""

    def __init__(self, addr):
        """Create an Address object."""
        self._addr = _normalize(addr)
        if self._addr is None:
            raise ValueError("Address cannot be None")

    def __repr__(self):
        """Representation of the Address object."""
        return self._addr.hex()

    def __str__(self):
        """Emit the address in human-readible format (AA.BB.CC)."""
        return "{}.{}.{}".format(
            self.__repr__()[0:2], self.__repr__()[2:4], self.__repr__()[4:6]
        ).upper()

    def __bytes__(self):
        """Return the bytes representation of the address."""
        return self._addr

    def __eq__(self, other):
        """Test for equality."""
        if isinstance(other, Address):
            return bytes(self) == bytes(other)
        return False

    def __ne__(self, other):
        """Test for not equals."""
        if isinstance(other, Address):
            return bytes(self) != bytes(other)
        return True

    def __lt__(self, other):
        """Test for less than."""
        if isinstance(other, Address):
            return bytes(self) < bytes(other)
        raise TypeError

    def __gt__(self, other):
        """Test for greater than."""
        if isinstance(other, Address):
            return bytes(self) > bytes(other)
        raise TypeError

    def __hash__(self):
        """Create a hash code for the Address object."""
        return hash(str(self))

    def __getitem__(self, byte):
        """Return a btye within the Address object."""
        if byte in [0, 1, 2]:
            return self._addr[byte]
        raise ValueError("Item index must be 0, 1 or 2")

    @property
    def id(self):
        """Return the address id."""
        return repr(self)

    @property
    def high(self):
        """High byte of the address."""
        return self._addr[0]

    @property
    def middle(self):
        """Middle byte of the address."""
        return self._addr[1]

    @property
    def low(self):
        """Low byte of the address."""
        return self._addr[2]



"""Termostat device types."""
import logging
from datetime import datetime
from ..aldb import ALDB
from ..aldb.aldb_battery import ALDBBattery
from ..default_link import DefaultLink
from ..groups import (
    COOL_SET_POINT,
    COOLING,
    DEHUMIDIFYING,
    FAN_MODE,
    HEAT_SET_POINT,
    HEATING,
    HUMIDIFYING,
    HUMIDITY_HIGH,
    HUMIDITY_LOW,
    SYSTEM_MODE,
    TEMPERATURE,
    HUMIDITY,
)
from ..extended_property import (
    BACKLIGHT,
    CHANGE_DELAY,
    TEMP_OFFSET,
    HUMIDITY_OFFSET,
)
from ..groups.thermostat import FanMode, Humidity, SystemMode, Temperature, SetPoint
from ..groups.on_off import OnOff
from ..managers.on_level_manager import OnLevelManager
from ..handlers.from_device.thermostat_status_response import (
    ThermostatStatusResponseHandler,
)
from ..handlers.from_device.thermostat_set_point_response import (
    ThermostatSetPointResponseHandler,
)
from ..handlers.from_device.thermostat_cool_set_point import (
    ThermostatCoolSetPointHandler,
)
from ..handlers.from_device.thermostat_heat_set_point import (
    ThermostatHeatSetPointHandler,
)
from ..handlers.from_device.thermostat_mode import ThermostatModeHandler
from ..handlers.from_device.thermostat_humidity import ThermostatHumidityHandler
from ..handlers.from_device.thermostat_temperature import ThermostatTemperatureHandler
from ..handlers.to_device.thermostat_cool_set_point import ThermostatCoolSetPointCommand
from ..handlers.to_device.thermostat_heat_set_point import ThermostatHeatSetPointCommand
from ..handlers.to_device.thermostat_mode import ThermostatModeCommand
from ..handlers.to_device.extended_set import ExtendedSetCommand
from ..handlers.to_device.extended_set_2 import ExtendedSet2Command
from ..managers.thermostat_status_manager import GetThermostatStatus
from ..operating_flag import (
    BUTTON_LOCK_ON,
    CELSIUS,
    KEY_BEEP_ON,
    LED_ON,
    PROGRAM_LOCK_ON,
    TIME_24_HOUR_FORMAT,
)
from .commands import STATUS_COMMAND
from .device_base import Device
from .battery_base import BatteryDeviceBase
from ..utils import multiple_status, to_fahrenheit, set_bit
from ..constants import ResponseStatus

_LOGGER = logging.getLogger(__name__)

GRP_COOL_ON = 1
GRP_HEAT_ON = 2
GRP_HUMID_HI_ON = 3
GRP_HUMID_LO_ON = 4
GRP_NOTIFY = 0xEF

GRP_TEMP = 10
GRP_HUMID = 11
GRP_SYS_MODE = 12
GRP_FAN_MODE = 13
GRP_COOL_SP = 14
GRP_HEAT_SP = 15
GRP_HUMID_HI_SP = 16
GRP_HUMID_LO_SP = 17

OP_FLAG_POS = 13


class ClimateControl_Thermostat(Device):
    """Thermostat device."""

    def __init__(self, address, cat, subcat, firmware=0x00, description="", model=""):
        """Init the Thermostat class."""
        super().__init__(
            address=address,
            cat=cat,
            subcat=subcat,
            firmware=firmware,
            description=description,
            model=model,
        )
        self._aldb = ALDB(self._address, mem_addr=0x1FFF)

    # pylint: disable=arguments-differ
    async def async_status(self, group=None):
        """Get the status of the device."""
        return await self._managers[STATUS_COMMAND].async_status()

    async def async_set_cool_set_point(self, temperature):
        """Set the cool set point."""
        temperature = max(1, min(temperature, 127))
        return await self._handlers["cool_set_point_command"].async_send(temperature)

    async def async_set_heat_set_point(self, temperature):
        """Set the cool set point."""
        temperature = max(1, min(temperature, 127))
        return await self._handlers["heat_set_point_command"].async_send(temperature)

    async def async_set_humidity_high_set_point(self, humidity):
        """Set the humdity high set point."""
        humidity = min(humidity, 99)
        cmd_status = await self._handlers["humidity_high_command"].async_send(humidity)
        if cmd_status == ResponseStatus.SUCCESS:
            self._groups[GRP_HUMID_HI_SP].value = humidity
        return cmd_status

    async def async_set_humidity_low_set_point(self, humidity):
        """Set the humidity low set point."""
        humidity = max(humidity, 1)
        cmd_status = await self._handlers["humidity_low_command"].async_send(humidity)
        if cmd_status == ResponseStatus.SUCCESS:
            self._groups[GRP_HUMID_LO_SP].value = humidity
        return cmd_status

    async def async_set_mode(self, mode):
        """Set the thermastat mode."""
        return await self._handlers["mode_command"].async_send(mode)

    async def async_set_master(self, master):
        """Set the thermastat master mode."""
        return await self._handlers["set_master"].async_send(master)

    async def async_set_notify_changes(self):
        """Set the thermostat to notify of changes."""
        return await self._handlers["notify_changes_command"].async_send()

    async def async_set_day_time(self):
        """Set the thermostat day of the week and time."""
        curr_time = datetime.now()
        day = 0 if curr_time.weekday() == 6 else curr_time.weekday() + 1
        set_time_command = ExtendedSet2Command(self._address, 0x02, day)
        return await set_time_command.async_send(
            data3=curr_time.hour, data4=curr_time.minute, data5=curr_time.second
        )

    async def async_add_default_links(self):
        """Add the default links betweent he modem and the device."""
        result_links = await super().async_add_default_links()
        result_notify = await self.async_set_notify_changes()
        return multiple_status(result_links, result_notify)

    async def async_write_op_flags(self):
        """Write the operating flags to the device."""
        flags = 0x00
        for flag_name in self._ext_property_manager.flag_info:
            flag_info = self._ext_property_manager.flag_info[flag_name]
            if flag_info.data_field == OP_FLAG_POS:
                flag = self.properties[flag_name]
                value = flag.new_value if flag.is_dirty else flag.value
                flags = set_bit(flags, flag_info.bit, bool(value))
        result = await self._handlers["op_flag_write"].async_send(data4=flags)
        if result == ResponseStatus.SUCCESS:
            for flag_name in self._ext_property_manager.flag_info:
                flag_info = self._ext_property_manager.flag_info[flag_name]
                if flag_info.data_field == OP_FLAG_POS:
                    flag = self.properties[flag_name]
                    if flag.is_dirty:
                        flag.load(flag.new_value)
        return result

    def _register_operating_flags(self):
        """Register thermostat operating flags."""
        self._add_property(PROGRAM_LOCK_ON, OP_FLAG_POS, None, 0, 0)
        self._add_property(KEY_BEEP_ON, OP_FLAG_POS, None, 0, 1)
        self._add_property(BUTTON_LOCK_ON, OP_FLAG_POS, None, 0, 2)
        self._add_property(CELSIUS, OP_FLAG_POS, None, 0, 3)
        self._add_property(TIME_24_HOUR_FORMAT, OP_FLAG_POS, None, 0, 4)
        self._add_property(LED_ON, OP_FLAG_POS, None, 0, 6)

        self._add_property(TEMP_OFFSET, 5, 2, 0)
        self._add_property(HUMIDITY_OFFSET, 6, 3, 0)
        self._add_property(BACKLIGHT, 9, 5, 0)
        self._add_property(CHANGE_DELAY, 10, 6, 0)

    def _register_groups(self):
        """Register the thermostat groups."""
        self._groups[GRP_COOL_ON] = OnOff(COOLING, self._address, GRP_COOL_ON)
        self._groups[GRP_HEAT_ON] = OnOff(HEATING, self._address, GRP_HEAT_ON)
        self._groups[GRP_HUMID_HI_ON] = OnOff(
            DEHUMIDIFYING, self._address, GRP_HUMID_HI_ON
        )
        self._groups[GRP_HUMID_LO_ON] = OnOff(
            HUMIDIFYING, self._address, GRP_HUMID_LO_ON
        )

        self._groups[GRP_TEMP] = Temperature(TEMPERATURE, self._address, GRP_TEMP, 0)
        self._groups[GRP_HUMID] = Humidity(
            HUMIDITY, self._address, group=GRP_HUMID, default=0
        )
        self._groups[GRP_SYS_MODE] = SystemMode(
            SYSTEM_MODE, self._address, group=GRP_SYS_MODE, default=0
        )
        self._groups[GRP_FAN_MODE] = FanMode(
            FAN_MODE, self._address, group=GRP_FAN_MODE, default=4
        )
        self._groups[GRP_COOL_SP] = SetPoint(
            COOL_SET_POINT, self._address, group=GRP_COOL_SP, default=65,
        )
        self._groups[GRP_HEAT_SP] = SetPoint(
            HEAT_SET_POINT, self._address, group=GRP_HEAT_SP, default=95,
        )
        self._groups[GRP_HUMID_HI_SP] = Humidity(
            HUMIDITY_HIGH, self._address, group=GRP_HUMID_HI_SP, default=0
        )
        self._groups[GRP_HUMID_LO_SP] = Humidity(
            HUMIDITY_LOW, self._address, group=GRP_HUMID_LO_SP, default=0
        )

    def _register_handlers_and_managers(self):
        """Register thermostat handlers and managers."""
        super()._register_handlers_and_managers()
        self._managers[GRP_COOL_ON] = OnLevelManager(self._address, GRP_COOL_ON, 0x00)
        self._managers[GRP_HEAT_ON] = OnLevelManager(self._address, GRP_HEAT_ON, 0x00)
        self._managers[GRP_HUMID_HI_ON] = OnLevelManager(
            self._address, GRP_HUMID_HI_ON, 0x00
        )
        self._managers[GRP_HUMID_LO_ON] = OnLevelManager(
            self._address, GRP_HUMID_LO_ON, 0x00
        )

        self._managers[STATUS_COMMAND] = GetThermostatStatus(self._address)
        self._handlers["status_response"] = ThermostatStatusResponseHandler(
            self._address
        )
        self._handlers["set_point_response"] = ThermostatSetPointResponseHandler(
            self._address
        )
        self._handlers["cool_set_point_handler"] = ThermostatCoolSetPointHandler(
            self._address
        )
        self._handlers["heat_set_point_handler"] = ThermostatHeatSetPointHandler(
            self._address
        )
        self._handlers["humidity_handler"] = ThermostatHumidityHandler(self._address)
        self._handlers["temperature_handler"] = ThermostatTemperatureHandler(
            self._address
        )
        self._handlers["mode_handler"] = ThermostatModeHandler(self._address)

        self._handlers["cool_set_point_command"] = ThermostatCoolSetPointCommand(
            self._address
        )
        self._handlers["heat_set_point_command"] = ThermostatHeatSetPointCommand(
            self._address
        )
        self._handlers["mode_command"] = ThermostatModeCommand(self._address)
        self._handlers["notify_changes_command"] = ExtendedSetCommand(
            self._address, 0x00, 0x08
        )
        self._handlers["humidity_high_command"] = ExtendedSetCommand(
            self._address, 0x00, 0x0B
        )
        self._handlers["humidity_low_command"] = ExtendedSetCommand(
            self._address, 0x00, 0x0C
        )
        self._handlers["set_master"] = ExtendedSetCommand(self._address, 0x00, 0x09)

        self._handlers["op_flag_write"] = ExtendedSetCommand(self._address, 0x00, 0x04)

    def _subscribe_to_handelers_and_managers(self):
        """Subscribe to handlers and managers."""
        super()._subscribe_to_handelers_and_managers()
        self._managers[GRP_COOL_ON].subscribe(self._groups[GRP_COOL_ON].set_value)
        self._managers[GRP_HEAT_ON].subscribe(self._groups[GRP_HEAT_ON].set_value)
        self._managers[GRP_HUMID_HI_ON].subscribe(
            self._groups[GRP_HUMID_HI_ON].set_value
        )
        self._managers[GRP_HUMID_LO_ON].subscribe(
            self._groups[GRP_HUMID_LO_ON].set_value
        )

        self._handlers["status_response"].subscribe(self._status_received)
        self._handlers["set_point_response"].subscribe(self._set_point_received)
        self._handlers["cool_set_point_handler"].subscribe(
            self._groups[GRP_COOL_SP].set_value
        )
        self._handlers["heat_set_point_handler"].subscribe(
            self._groups[GRP_HEAT_SP].set_value
        )
        self._handlers["humidity_handler"].subscribe(self._groups[GRP_HUMID].set_value)
        self._handlers["temperature_handler"].subscribe(self._temp_received)
        self._handlers["mode_handler"].subscribe(self._mode_received)
        self._properties[CELSIUS].subscribe(self._temp_format_first_set)

    def _register_default_links(self):
        """Register default links."""
        super()._register_default_links()
        link_1 = DefaultLink(
            is_controller=True,
            group=GRP_COOL_ON,
            dev_data1=0,
            dev_data2=0,
            dev_data3=GRP_COOL_ON,
            modem_data1=0,
            modem_data2=0,
            modem_data3=GRP_COOL_ON,
        )
        link_2 = DefaultLink(
            is_controller=True,
            group=GRP_HEAT_ON,
            dev_data1=0,
            dev_data2=0,
            dev_data3=GRP_HEAT_ON,
            modem_data1=0,
            modem_data2=0,
            modem_data3=GRP_HEAT_ON,
        )
        link_3 = DefaultLink(
            is_controller=True,
            group=GRP_HUMID_HI_ON,
            dev_data1=0,
            dev_data2=0,
            dev_data3=GRP_HUMID_HI_ON,
            modem_data1=0,
            modem_data2=0,
            modem_data3=GRP_HUMID_HI_ON,
        )
        link_4 = DefaultLink(
            is_controller=True,
            group=GRP_HUMID_LO_ON,
            dev_data1=0,
            dev_data2=0,
            dev_data3=GRP_HUMID_LO_ON,
            modem_data1=0,
            modem_data2=0,
            modem_data3=GRP_HUMID_LO_ON,
        )
        link_ef = DefaultLink(
            is_controller=True,
            group=GRP_NOTIFY,
            dev_data1=0x03,
            dev_data2=0,
            dev_data3=GRP_NOTIFY,
            modem_data1=0,
            modem_data2=0,
            modem_data3=GRP_NOTIFY,
        )
        self._default_links.append(link_1)
        self._default_links.append(link_2)
        self._default_links.append(link_3)
        self._default_links.append(link_4)
        self._default_links.append(link_ef)

    def _status_received(
        self,
        day,
        hour,
        minute,
        second,
        system_mode,
        fan_mode,
        cool_set_point,
        humidity,
        temperature,
        cooling,
        heating,
        celsius,
        heat_set_point,
    ):
        """Receive the status update."""
        self._groups[GRP_COOL_ON].set_value(cooling)
        self._groups[GRP_HEAT_ON].set_value(heating)
        self._groups[GRP_SYS_MODE].set_value(system_mode)
        self._groups[GRP_FAN_MODE].set_value(fan_mode)
        self._groups[GRP_COOL_SP].set_value(cool_set_point)
        self._groups[GRP_HEAT_SP].set_value(heat_set_point)
        if not self._properties[CELSIUS].value:
            temperature = to_fahrenheit(temperature)
        self._groups[GRP_TEMP].set_value(temperature)
        self._groups[GRP_HUMID].set_value(humidity)

    def _set_point_received(
        self,
        stage_1_on_minutes,
        humidity_high,
        humidity_low,
        firmwire,
        cool_set_point,
        heat_set_point,
        rf_offset,
    ):
        """Receive set point info."""
        self._groups[GRP_COOL_SP].set_value(cool_set_point)
        self._groups[GRP_HEAT_SP].set_value(heat_set_point)
        self._groups[GRP_HUMID_HI_SP].set_value(humidity_high)
        self._groups[GRP_HUMID_LO_SP].set_value(humidity_low)

    def _mode_received(self, system_mode, fan_mode):
        """Receive current temperature notification."""
        self._groups[GRP_SYS_MODE].set_value(system_mode)
        self._groups[GRP_FAN_MODE].set_value(fan_mode)

    def _temp_received(self, degrees):
        """Receive temperature status update and convert to celsius if needed."""
        self._groups[GRP_TEMP].value = degrees

    def _temp_format_changed(self, name, value):
        """Recieve notification that the thermostat has changed to/from C/F."""
        self.status()

    def _temp_format_first_set(self, name, value):
        """Set up the trigger for a temperature format change.

        The first time the format is set, we don't need to do anything. If
        the format changes later, we need to get the status update to change the
        measurements from F to C or vise versa.
        """
        self._properties[CELSIUS].unsubscribe(self._temp_format_first_set)
        self._properties[CELSIUS].unsubscribe(self._temp_format_changed)


class ClimateControl_WirelessThermostat(BatteryDeviceBase, ClimateControl_Thermostat):
    """Wireless Thermostat device."""

    def __init__(self, address, cat, subcat, firmware=0x00, description="", model=""):
        """Init the Wireless Thermostat class."""
        super(ClimateControl_WirelessThermostat, self).__init__(
            address=address,
            cat=cat,
            subcat=subcat,
            firmware=firmware,
            description=description,
            model=model,
        )
        self._aldb = ALDBBattery(
            address=address, mem_addr=0x1FFF, run_command=self._run_on_wake
        )

    async def async_set_humidity_high_set_point(self, humidity):
        """Set the humdity high set point."""
        return self._run_on_wake(
            super(BatteryDeviceBase, self).async_set_humidity_high_set_point,
            humidity=humidity,
        )

    async def async_set_humidity_low_set_point(self, humidity):
        """Set the humidity low set point."""
        return self._run_on_wake(
            super(BatteryDeviceBase, self).async_set_humidity_low_set_point,
            humidity=humidity,
        )

    async def async_set_master(self, master):
        """Set the thermastat master mode."""
        return self._run_on_wake(
            super(BatteryDeviceBase, self).async_set_master, master=master
        )

    async def async_set_mode(self, mode):
        """Set the thermastat mode."""
        return self._run_on_wake(
            super(BatteryDeviceBase, self).async_set_mode, mode=mode
        )

    async def async_set_notify_changes(self):
        """Set the thermostat to notify of changes."""
        return self._run_on_wake(
            super(BatteryDeviceBase, self).async_set_notify_changes
        )

    async def async_set_cool_set_point(self, temperature):
        """Set the cool set point."""
        return self._run_on_wake(
            super(BatteryDeviceBase, self).async_set_cool_set_point,
            temperature=temperature,
        )

    async def async_set_day_time(self):
        """Set the thermostat day of the week and time."""
        return self._run_on_wake(super(BatteryDeviceBase, self).async_set_day_time)

    async def async_set_heat_set_point(self, temperature):
        """Set the cool set point."""
        return self._run_on_wake(
            super(BatteryDeviceBase, self).async_set_heat_set_point,
            temperature=temperature,
        )



"""All-Link database for battery oppertated devices."""
import logging
from collections import namedtuple
from typing import Callable

from ..constants import ALDBVersion
from . import ALDB

LoadCommandParams = namedtuple(
    "LoadCommandParams", "mem_addr num_recs refresh callback"
)
Command = namedtuple("Command", "action params")

_LOGGER = logging.getLogger(__name__)


class ALDBBattery(ALDB):
    """ALDB for battery opperated devices."""

    def __init__(
        self, address, version=ALDBVersion.V2, mem_addr=0x0FFF, run_command=None
    ):
        """Init the ALDBBattery class."""
        super().__init__(address=address, version=version, mem_addr=mem_addr)
        self._commands = []
        self._run_command = run_command

    # pylint: disable=arguments-differ
    async def async_load(
        self,
        mem_addr: int = 0x00,
        num_recs: int = 0x00,
        refresh: bool = False,
        callback: Callable = None,
    ):
        """Load the All-Link Database."""
        self._run_command(
            super().async_load,
            mem_addr=mem_addr,
            num_recs=num_recs,
            refresh=refresh,
            callback=callback,
        )
        return True

    async def async_write(self, force=False):
        """Write modified records to the device."""
        self._run_command(super().async_write, force=force)
        return 0, 0



"""Open / Close sensor groups."""
from .group_base import GroupBase
from ..address import Address


class NormallyOpen(GroupBase):
    """Normally open sensor state."""

    def __init__(
        self, name: str, address: Address, group: int = 0, default: int = None
    ):
        """Init the OnLevel class."""
        super().__init__(name, address, group, default, value_type=int)

    # pylint: disable=arguments-differ
    def set_value(self, on_level):
        """Set the value of the state from the handlers."""
        # Off is a Open state and On is an Closed state
        self.value = 0xFF if on_level else 0


class NormallyClosed(GroupBase):
    """Normally closed sensor state."""

    def __init__(
        self, name: str, address: Address, group: int = 0, default: int = None
    ):
        """Init the OnLevel class."""
        super().__init__(name, address, group, default, value_type=int)

    # pylint: disable=arguments-differ
    def set_value(self, on_level):
        """Set the value of the state from the handlers."""
        # Off is a Closed state and On is an Open state
        self.value = 0 if on_level else 0xFF



"""All-Link Database for devices with no All-Link Database."""
from typing import Callable

from ..address import Address
from ..constants import ALDBStatus, ALDBVersion


class NoALDB:
    """All-Link Database for devices with no All-Link Database."""

    def __init__(self, address, version=ALDBVersion.V2, mem_addr=0x0FFF):
        """Do nothing."""
        self._address = address

    def __setitem__(self, mem_addr, record):
        """Do nothing."""

    def __iter__(self):
        """Iterate through each ALDB device record."""
        for addr in {}:
            yield addr

    def __getitem__(self, mem_addr):
        """Fetch a device from the ALDB."""
        return {}

    def __len__(self):
        """Return the number of devices in the ALDB."""
        return 0

    def __repr__(self):
        """Human representation of a device from the ALDB."""
        attrs = vars(self)
        return ", ".join("%s: %r" % item for item in attrs.items())

    @property
    def is_loaded(self) -> bool:
        """Return loaded always."""
        return True

    @property
    def status(self) -> ALDBStatus:
        """Return loaded status."""
        return ALDBStatus.LOADED

    # pylint: disable=arguments-differ
    async def async_load(
        self,
        mem_addr: int = 0x00,
        num_recs: int = 0x00,
        refresh: bool = False,
        callback: Callable = None,
    ):
        """Load the All-Link Database."""
        if callback:
            callback()
        return ALDBStatus.LOADED

    async def async_write(self):
        """Write modified records to the device."""
        return 0, 0

    def add(
        self,
        group: int,
        target: Address,
        controller: bool = False,
        data1: int = 0x00,
        data2: int = 0x00,
        data3: int = 0x00,
    ):
        """Add an All-Link record."""

    def remove(self, mem_addr: int):
        """Remove an All-Link record."""

    @property
    def address(self):
        """Returnt the status of the device."""
        return self._address

    @property
    def version(self) -> ALDBVersion:
        """Return the ALDB version."""
        return ALDBVersion.NULL

    @property
    def first_mem_addr(self):
        """Return the expected memory address of the first record."""
        return 0x0FFF

    @property
    def high_water_mark_mem_addr(self):
        """Return the High Water Mark record memory address."""
        return None

    # pylint: disable=no-self-use
    def get(self, mem_addr, default=None):
        """Get the record at address 'mem_addr'."""
        return None

    # pylint: disable=no-self-use
    def get_responders(self, group):
        """Return all responders to this device for a group."""
        return None

    def update_version(self, version):
        """Update the ALDB version number."""

    def load_saved_records(self, status, records):
        """Load All-Link records from a dictionary of saved records."""



"""Insteon Modem Base Class."""
import asyncio
from abc import ABCMeta
from asyncio import Transport

from ..handlers.get_im_configuration import GetImConfigurationHandler
from ..handlers.set_im_configuration import SetImConfigurationHandler
from ..handlers.all_link_cleanup_report import AllLinkCleanupStatusReport
from ..handlers.all_link_cleanup_failure_report import AllLinkCleanupFailureReport
from ..protocol.protocol import Protocol
from .commands import GET_IM_CONFIG_COMMAND
from .device_base import Device


class ModemBase(Device, metaclass=ABCMeta):
    """Base class for insteon Modems (PLM and Hub)."""

    __meta__ = ABCMeta

    def __init__(
        self,
        address="000000",
        cat=0x03,
        subcat=0x00,
        firmware=0x00,
        description="",
        model="",
    ):
        """Init the Modem class."""
        super().__init__(address, cat, subcat, firmware, description, model)
        self._aldb = None
        self._subscribe_topics()
        self._protocol = None
        self._transport = None
        self._disable_auto_linking = False
        self._monitor_mode = False
        self._auto_led = False
        self._deadman = False

    @property
    def connected(self) -> bool:
        """Return true if the transport is connected."""
        if not self._protocol:
            return False
        return self._protocol.connected

    @property
    def protocol(self):
        """Return the protocol."""
        return self._protocol

    @property
    def disable_auto_linking(self):
        """Return the Disable Auto Linking flag value."""
        self._disable_auto_linking = False

    @property
    def monitor_mode(self):
        """Return the Monitor Mode flag value."""
        self._monitor_mode = False

    @property
    def auto_led(self):
        """Return the Auto LED flag value."""
        self._auto_led = False

    @property
    def deadman(self):
        """Return the Deadman flag value."""
        self._deadman = False

    @protocol.setter
    def protocol(self, value):
        """Set the protocol."""
        if isinstance(value, Protocol):
            self._protocol = value

    @property
    def transport(self):
        """Return the transport."""
        return self._transport

    @transport.setter
    def transport(self, value):
        """Set the transport."""
        if isinstance(value, Transport):
            self._transport = value

    def close(self):
        """Close the connection to the transport."""
        asyncio.ensure_future(self.async_close())

    async def async_close(self):
        """Close the connection to the transport ascynronously."""
        if self._protocol:
            self._protocol.close()
            wait_time = 0.0001
            while self.connected:
                await asyncio.sleep(wait_time)
                wait_time = min(300, 1.5 * wait_time)

    async def async_get_configuration(self):
        """Get the modem flags."""
        return await self._handlers[GET_IM_CONFIG_COMMAND].async_send()

    def _update_flags(
        self,
        disable_auto_linking: bool,
        monitor_mode: bool,
        auto_led: bool,
        deadman: bool,
    ):
        self._disable_auto_linking = disable_auto_linking
        self._monitor_mode = monitor_mode
        self._auto_led = auto_led
        self._deadman = deadman

    async def async_set_configuration(
        self,
        disable_auto_linking: bool,
        monitor_mode: bool,
        auto_led: bool,
        deadman: bool,
    ):
        """Set the modem flags."""
        return await SetImConfigurationHandler().async_send(
            disable_auto_linking=disable_auto_linking,
            monitor_mode=monitor_mode,
            auto_led=auto_led,
            deadman=deadman,
        )

    async def async_get_operating_flags(self, group=None):
        """Read the device operating flags."""

    async def async_set_operating_flags(self, group=None, force=False):
        """Write the operating flags to the device."""

    async def async_get_extended_properties(self, group=None):
        """Get the device extended properties."""

    def _subscribe_topics(self):
        """Subscribe to modem specific topics."""

    def _register_groups(self):
        """No groups to register for modems."""

    def _register_default_links(self):
        """No default links for modems."""

    def _register_handlers_and_managers(self):
        """Register command handlers for modems."""
        super()._register_handlers_and_managers()
        self._handlers["cleanup_status_report"] = AllLinkCleanupStatusReport()
        self._handlers["cleanup_failure_report"] = AllLinkCleanupFailureReport()
        self._handlers[GET_IM_CONFIG_COMMAND] = GetImConfigurationHandler()
        self._handlers[GET_IM_CONFIG_COMMAND].subscribe(self._update_flags)

    def _register_events(self):
        """Register events for modems."""

    def _register_operating_flags(self):
        """Register operating flags for modem."""



"""KeypadLinc command handler to trigger a button scene."""
from .. import ack_handler, direct_ack_handler
from ...topics import EXTENDED_TRIGGER_ALL_LINK
from ..to_device.direct_command import DirectCommandHandlerBase


class TriggerSceneOnCommandHandler(DirectCommandHandlerBase):
    """KeypadLinc command handler to trigger a button scene."""

    def __init__(self, address, group):
        """Init the SetLedCommandHandler class."""
        super().__init__(topic=EXTENDED_TRIGGER_ALL_LINK, address=address, group=group)
        self._on_level = 0

    # pylint: disable=arguments-differ
    async def async_send(self, on_level: int = 0xFF, fast: bool = False):
        """Set the LED values of the KPL."""
        self._on_level = on_level
        action = 0 if on_level == 0xFF else 1
        cmd2 = self._on_level
        ramp_rate = 1 if fast else 0
        kwargs = {
            "data1": self._group,
            "data2": action,
            "data3": self._on_level,
            "data4": 0x11,
            "data5": cmd2,
            "data6": ramp_rate,
        }

        return await super().async_send(**kwargs)

    @ack_handler(True)
    def handle_ack(self, cmd1, cmd2, user_data):
        """Handle the ACK response.

        Required since the BEEP command uses the same cmd1, cmd2 values.
        """
        if not user_data and not user_data["data1"] == self._group:
            return
        return super().handle_ack(cmd1, cmd2, user_data)

    @direct_ack_handler
    def handle_direct_ack(self, cmd1, cmd2, target, user_data, hops_left):
        """Handle the direct ACK message."""
        self._call_subscribers(on_level=self._on_level)



"""Device manager."""
import logging
import asyncio

from ..address import Address
from ..x10_address import X10Address
from ..device_types.device_base import Device
from ..device_types.x10_base import X10DeviceBase
from ..device_types.modem_base import ModemBase
from ..managers.saved_devices_manager import SavedDeviceManager
from ..subscriber_base import SubscriberBase
from .device_id_manager import DeviceId, DeviceIdManager
from .device_link_manager import DeviceLinkManager
from .utils import create_device, create_x10_device

DEVICE_INFO_FILE = "insteon_devices.json"
_LOGGER = logging.getLogger(__name__)


# TODO remove devices
class DeviceManager(SubscriberBase):
    """Manages the list of active devices."""

    def __init__(self):
        """Init the DeviceManager class."""
        super().__init__(subscriber_topic="device_added")
        self._devices = {}
        self._modem = None
        self._id_manager = DeviceIdManager()
        self._id_manager.subscribe(self._device_identified)
        self._loading_saved_lock = asyncio.Lock()
        self._link_manager = DeviceLinkManager(self)

    def __getitem__(self, address) -> Device:
        """Return a a device from the device address."""
        try:
            address = Address(address)
        except ValueError:
            address = X10Address(address)
        return self._devices.get(address)

    def __iter__(self):
        """Return an iterator of device addresses."""
        for address in self._devices:
            yield address

    def __setitem__(self, address, device):
        """Add a device to the device list."""
        _LOGGER.info("Adding device to INSTEON devices list: %s", address.id)
        if not isinstance(device, (Device, DeviceId, X10DeviceBase)):
            raise ValueError("Device must be a DeviceId or a Device type.")

        if isinstance(device, DeviceId):
            device = create_device(device)

        self._devices[device.address] = device
        if isinstance(device, Device):
            self._id_manager.set_device_id(
                device.address, device.cat, device.subcat, device.firmware
            )
        self._call_subscribers(address=device.address.id)

    def __len__(self):
        """Return the number of devices."""
        return len(self._devices)

    def get(self, address) -> Device:
        """Return a device from an address."""
        try:
            address = Address(address)
        except ValueError:
            address = X10Address(address)
        return self._devices.get(address)

    def pop(self, address):
        """Remove a device from the device list."""
        try:
            address = Address(Address)
        except ValueError:
            address = X10Address(address)
        self._devices.pop(address)

    @property
    def modem(self):
        """Return the Insteon Modem."""
        return self._modem

    @modem.setter
    def modem(self, modem):
        """Set the Insteon Modem."""
        if not isinstance(modem, ModemBase):
            raise ValueError("Must be an Insteon Modem object")
        self._modem = modem
        self._devices[self._modem.address] = self._modem

    @property
    def id_manager(self):
        """Return the ID manager instance."""
        return self._id_manager

    def set_id(self, address: Address, cat: int, subcat: int, firmware: int):
        """Add a device override to identify the device information.

        Typical use is to identify devices that do not respond to an ID Request
        such as a battery opperated device.

        """
        address = Address(address)
        device = self[address]
        if device and device.cat == cat and device.subcat == subcat:
            return
        self._id_manager.set_device_id(address, cat, subcat, firmware)

    async def async_identify_device(self, address: Address):
        """Identify a device.

        The device will be placed into the unknown device list to be identified.

        If the device has already been identified, this method will remove the device
        from the known device list. This is typically used when a `set_id` command has
        been run to create a device override. The `async_reidentify_device` command will
        reset that override and allow normal device identification to run.
        """
        self._devices.pop(Address(address))
        await self._id_manager.async_id_device(address=address, refresh=True)

    def add_x10_device(
        self,
        housecode: str,
        unitcode: int,
        x10_feature: str,
        steps: int = 22,
        max_level: int = 255,
    ):
        """Add an X10 device."""
        device = create_x10_device(housecode, unitcode, x10_feature, steps, max_level)
        if device:
            self[device.address] = device
        return device

    async def async_close(self):
        """Close the device ID listener."""
        self._id_manager.close()

    async def async_load(self, workdir="", id_devices=1, load_modem_aldb=1):
        """Load devices from the `insteon_devices.yaml` file and device overrides.

        Parameters:
            workdir: Directory name to find the `insteon_devices.json` file

            id_devices: Indicate if devices should be identified using ID Request
                0: No devices are identified
                1: Unknown devices are identified
                2: All devices are identified
            (default=1)

            load_modem_aldb: Indicate if the Modem ALDB should be loaded
                0: Do not load
                1: Load if not loaded from save file
                2: Load
            (default=1)

        The Modem ALDB is loaded if `refresh` is True or if the saved file has no devices.

        """
        if workdir:
            async with self._loading_saved_lock:
                saved_devices_manager = SavedDeviceManager(workdir, self.modem)
                devices = await saved_devices_manager.async_load()
                for address in devices:
                    self[address] = devices[address]

        if load_modem_aldb == 0:
            load_modem_aldb = False
        elif load_modem_aldb == 2:
            load_modem_aldb = True
        else:
            load_modem_aldb = not self._modem.aldb.is_loaded

        if load_modem_aldb:
            await self._modem.aldb.async_load()

        for mem_addr in self._modem.aldb:
            rec = self._modem.aldb[mem_addr]
            if rec.target != Address("000000"):
                self._id_manager.append(rec.target)

        if id_devices:
            id_all = id_devices == 2
            await self._id_manager.async_id_devices(refresh=id_all)

    async def async_save(self, workdir):
        """Save devices to a device information file."""
        saved_devices_manager = SavedDeviceManager(workdir, self.modem)
        await saved_devices_manager.async_save(self._devices)

    def _device_identified(self, device_id: DeviceId):
        """Device identified by device ID manager."""
        if self._loading_saved_lock.locked():
            return
        if device_id.cat is not None:
            device = create_device(device_id)
            if self[device_id.address]:
                # If the device is already in the devices list and the cat and subcat
                # are the same, do not add the device again
                if (
                    device_id.cat == self[device_id.address].cat
                    and device_id.subcat == self._devices[device_id.address].subcat
                ):
                    return
            self[device_id.address] = device
            if device_id.cat != 0x03:
                asyncio.ensure_future(device.async_get_engine_version())
                asyncio.ensure_future(self.async_setup_device(device))
            _LOGGER.debug("Device %s added", device.address)

    async def async_setup_device(self, device):
        """Set up device."""
        await device.aldb.async_load(refresh=True)
        await device.async_read_op_flags()
        await device.async_read_ext_properties()
        await device.async_add_default_links()
